"""Base data backend for Ralph."""

import logging
from abc import ABC, abstractmethod
from asyncio import Queue, create_task
from enum import Enum, IntEnum, unique
from inspect import isclass
from io import IOBase
from itertools import chain
from typing import (
    Any,
    AsyncIterator,
    Generic,
    Iterable,
    Iterator,
    Optional,
    Set,
    Type,
    TypeVar,
    Union,
    get_args,
    get_origin,
)

from pydantic import BaseModel, BaseSettings, PositiveInt, ValidationError
from typing_extensions import Self, get_original_bases

from ralph.conf import BaseSettingsConfig, core_settings
from ralph.exceptions import BackendParameterException
from ralph.utils import (
    async_parse_dict_to_bytes,
    async_parse_iterable_to_dict,
    gather_with_limited_concurrency,
    iter_by_batch,
    parse_dict_to_bytes,
    parse_iterable_to_dict,
)

logger = logging.getLogger(__name__)


class BaseDataBackendSettings(BaseSettings):
    """Data backend default configuration."""

    class Config(BaseSettingsConfig):
        """Pydantic Configuration."""

        env_prefix = "RALPH_BACKENDS__DATA__"
        env_file = ".env"
        env_file_encoding = core_settings.LOCALE_ENCODING

    LOCALE_ENCODING: str = "utf8"
    READ_CHUNK_SIZE: int = 500
    WRITE_CHUNK_SIZE: int = 500


class BaseQuery(BaseModel):
    """Base query model."""

    class Config:
        """Base query model configuration."""

        extra = "forbid"

    @classmethod
    def from_string(cls, query: str) -> Self:
        """Return an instance of BaseQuery from a string."""
        try:
            return cls.parse_raw(query)
        except ValidationError as error:
            msg = "Invalid %s query string: %s"
            errors = error.errors()
            logger.error(msg, cls.__name__, errors)
            raise BackendParameterException(msg % (cls.__name__, errors)) from error


@unique
class BaseOperationType(Enum):
    """Base data backend operation types.

    Attributes:
        INDEX (str): creates a new record with a specific ID.
        CREATE (str): creates a new record without a specific ID.
        DELETE (str): deletes an existing record.
        UPDATE (str): updates or overwrites an existing record.
        APPEND (str): creates or appends data to an existing record.
    """

    INDEX = "index"
    CREATE = "create"
    DELETE = "delete"
    UPDATE = "update"
    APPEND = "append"


@unique
class DataBackendStatus(Enum):
    """Data backend statuses."""

    OK = "ok"
    AWAY = "away"
    ERROR = "error"


class Configurable:
    """A class that is configurable via data backend settings."""

    settings: BaseDataBackendSettings


class Writable(Configurable, ABC):
    """Data backend interface for backends supporting the write operation."""

    default_operation_type = BaseOperationType.INDEX
    unsupported_operation_types: Set[BaseOperationType] = set()

    def write(  # noqa: PLR0913
        self,
        data: Union[IOBase, Iterable[bytes], Iterable[dict]],
        target: Optional[str] = None,
        chunk_size: Optional[int] = None,
        ignore_errors: bool = False,
        operation_type: Optional[BaseOperationType] = None,
    ) -> int:
        """Write `data` records to the `target` container and return their count.

        Args:
            data (Iterable or IOBase): The data to write.
            target (str or None): The target container name.
                If `target` is `None`, a default value is used instead.
            chunk_size (int or None): The number of records or bytes to write in one
                batch, depending on whether `data` contains dictionaries or bytes.
                If `chunk_size` is `None` it defaults to `WRITE_CHUNK_SIZE`.
            ignore_errors (bool): If `True`, escapable errors are ignored and logged.
                If `False` (default), a `BackendException` is raised on any error.
            operation_type (BaseOperationType or None): The mode of the write operation.
                If `operation_type` is `None`, the `default_operation_type` is used
                instead. See `BaseOperationType`.

        Return:
            int: The number of written records.

        Raise:
            BackendException: If any failure occurs during the write operation or
                if an inescapable failure occurs and `ignore_errors` is set to `True`.
            BackendParameterException: If a backend argument value is not valid.
        """
        if not operation_type:
            operation_type = self.default_operation_type

        if operation_type in self.unsupported_operation_types:
            msg = f"{operation_type.value.capitalize()} operation_type is not allowed"
            logger.error(msg)
            raise BackendParameterException(msg)

        data = iter(data)
        try:
            first_record = next(data)
        except StopIteration:
            logger.info("Data Iterator is empty; skipping write to target")
            return 0
        data = chain((first_record,), data)

        chunk_size = chunk_size if chunk_size else self.settings.WRITE_CHUNK_SIZE
        is_bytes = isinstance(first_record, bytes)
        writer = self._write_bytes if is_bytes else self._write_dicts
        return writer(data, target, chunk_size, ignore_errors, operation_type)

    def _write_bytes(  # noqa: PLR0913
        self,
        data: Iterable[bytes],
        target: Optional[str],
        chunk_size: int,
        ignore_errors: bool,
        operation_type: BaseOperationType,
    ) -> int:
        """Method called by `self.write` writing bytes. See `self.write`."""
        statements = parse_iterable_to_dict(data, ignore_errors)
        return self._write_dicts(
            statements, target, chunk_size, ignore_errors, operation_type
        )

    @abstractmethod
    def _write_dicts(  # noqa: PLR0913
        self,
        data: Iterable[dict],
        target: Optional[str],
        chunk_size: int,
        ignore_errors: bool,
        operation_type: BaseOperationType,
    ) -> int:
        """Method called by `self.write` writing dictionaries. See `self.write`."""
        locale = self.settings.LOCALE_ENCODING
        statements = parse_dict_to_bytes(data, locale, ignore_errors)
        return self._write_bytes(
            statements, target, chunk_size, ignore_errors, operation_type
        )


class Listable(ABC):
    """Data backend interface for backends supporting the list operation."""

    @abstractmethod
    def list(
        self, target: Optional[str] = None, details: bool = False, new: bool = False
    ) -> Union[Iterator[str], Iterator[dict]]:
        """List containers in the data backend. E.g., collections, files, indexes.

        Args:
            target (str or None): The target container name.
                If `target` is `None`, a default value is used instead.
            details (bool): Get detailed container information instead of just names.
            new (bool): Given the history, list only not already read containers.

        Yield:
            str: If `details` is False.
            dict: If `details` is True.

        Raise:
            BackendException: If a failure occurs.
            BackendParameterException: If a backend argument value is not valid.
        """


Settings = TypeVar("Settings", bound=BaseDataBackendSettings)
Query = TypeVar("Query", bound=Union[BaseQuery, str])


def validate_backend_query(
    query: Optional[Query],
    query_class: Type[Query],
) -> Query:
    """Validate and transform the backend query."""
    query_name = query_class.__name__
    if query is None:
        try:
            return query_class()
        except ValidationError as error:
            msg = "Invalid %s default query: %s"
            errors = error.errors()
            logger.error(msg, query_name, errors)
            raise BackendParameterException(msg % (query_name, errors)) from error

    if isinstance(query, query_class):
        return query

    msg = "The 'query' argument is expected to be a %s instance"
    logger.error(msg, query_name)
    raise BackendParameterException(msg % query_name)


class DataBackendArgument(IntEnum):
    """Enumerate data backend generic arguments."""

    SETTINGS = 0
    QUERY = 1


class BaseDataBackend(Generic[Settings, Query], ABC):
    """Base data backend interface."""

    name = "base"
    query_class: Type[Query]
    settings_class: Type[Settings]

    def __init_subclass__(cls, **kwargs: Any) -> None:  # noqa: D105
        super().__init_subclass__(**kwargs)
        set_backend_settings_class(cls)
        set_backend_query_class(cls)

    def __init__(self, settings: Optional[Settings] = None):
        """Instantiate the data backend.

        Args:
            settings (Settings or None): The data backend settings.
                If `settings` is `None`, a default settings instance is used instead.
        """
        try:
            self.settings: Settings = settings if settings else self.settings_class()
        except ValidationError as error:
            msg = "Failed to instantiate default data backend settings: %s"
            logger.error(msg, error)
            raise BackendParameterException(msg % error) from error

    @abstractmethod
    def status(self) -> DataBackendStatus:
        """Implement data backend checks (e.g. connection, cluster status).

        Return:
            DataBackendStatus: The status of the data backend.
        """

    def read(  # noqa: PLR0913
        self,
        query: Optional[Query] = None,
        target: Optional[str] = None,
        chunk_size: Optional[int] = None,
        raw_output: bool = False,
        ignore_errors: bool = False,
        max_statements: Optional[PositiveInt] = None,
    ) -> Union[Iterator[bytes], Iterator[dict]]:
        """Read records matching the `query` in the `target` container and yield them.

        Args:
            query (Query): The query to select records to read.
            target (str or None): The target container name.
                If `target` is `None`, a default value is used instead.
            chunk_size (int or None): The number of records or bytes to read in one
                batch, depending on whether the records are dictionaries or bytes.
                If `chunk_size` is `None` it defaults to `READ_CHUNK_SIZE`.
            raw_output (bool): Controls whether to yield bytes or dictionaries.
                If the records are dictionaries and `raw_output` is set to `True`, they
                are encoded as JSON.
                If the records are bytes and `raw_output` is set to `False`, they are
                decoded as JSON by line.
            ignore_errors (bool): If `True`, encoding errors during the read operation
                will be ignored and logged.
                If `False` (default), a `BackendException` is raised on any error.
            max_statements (int): The maximum number of statements to yield.
                If `None` (default) or `0`, there is no maximum.

        Yield:
            dict: If `raw_output` is False.
            bytes: If `raw_output` is True.

        Raise:
            BackendException: If a failure during the read operation occurs or
                during encoding records and `ignore_errors` is set to `False`.
            BackendParameterException: If a backend argument value is not valid.
        """
        chunk_size = chunk_size if chunk_size else self.settings.READ_CHUNK_SIZE
        query = validate_backend_query(query, self.query_class)
        reader = self._read_bytes if raw_output else self._read_dicts
        statements = reader(query, target, chunk_size, ignore_errors)
        if not max_statements:
            yield from statements
            return

        max_statements -= 1
        for i, statement in enumerate(statements):
            yield statement
            if i >= max_statements:
                return

    def _read_bytes(
        self, query: Query, target: Optional[str], chunk_size: int, ignore_errors: bool
    ) -> Iterator[bytes]:
        """Method called by `self.read` yielding bytes. See `self.read`."""
        locale = self.settings.LOCALE_ENCODING
        statements = self._read_dicts(query, target, chunk_size, ignore_errors)
        yield from parse_dict_to_bytes(statements, locale, ignore_errors)

    @abstractmethod
    def _read_dicts(
        self, query: Query, target: Optional[str], chunk_size: int, ignore_errors: bool
    ) -> Iterator[dict]:
        """Method called by `self.read` yielding dictionaries. See `self.read`."""
        statements = self._read_bytes(query, target, chunk_size, ignore_errors)
        yield from parse_iterable_to_dict(statements, ignore_errors)

    @abstractmethod
    def close(self) -> None:
        """Close the data backend client.

        Raise:
            BackendException: If a failure occurs during the close operation.
        """


class AsyncWritable(Configurable, ABC):
    """Async data backend interface for backends supporting the write operation."""

    default_operation_type = BaseOperationType.INDEX
    unsupported_operation_types: Set[BaseOperationType] = set()

    async def write(  # noqa: PLR0913
        self,
        data: Union[IOBase, Iterable[bytes], Iterable[dict]],
        target: Optional[str] = None,
        chunk_size: Optional[int] = None,
        ignore_errors: bool = False,
        operation_type: Optional[BaseOperationType] = None,
        concurrency: Optional[PositiveInt] = None,
    ) -> int:
        """Write `data` records to the `target` container and return their count.

        Args:
            data (Iterable or IOBase): The data to write.
            target (str or None): The target container name.
                If `target` is `None`, a default value is used instead.
            chunk_size (int or None): The number of records or bytes to write in one
                batch, depending on whether `data` contains dictionaries or bytes.
                If `chunk_size` is `None` it defaults to `WRITE_CHUNK_SIZE`.
            ignore_errors (bool): If `True`, escapable errors are ignored and logged.
                If `False` (default), a `BackendException` is raised on any error.
            operation_type (BaseOperationType or None): The mode of the write operation.
                If `operation_type` is `None`, the `default_operation_type` is used
                instead. See `BaseOperationType`.
            concurrency (int): The number of chunks to write concurrently.
                If `None` it defaults to `1`.

        Return:
            int: The number of written records.

        Raise:
            BackendException: If any failure occurs during the write operation or
                if an inescapable failure occurs and `ignore_errors` is set to `True`.
            BackendParameterException: If a backend argument value is not valid.
        """
        if not operation_type:
            operation_type = self.default_operation_type

        if operation_type in self.unsupported_operation_types:
            msg = f"{operation_type.value.capitalize()} operation_type is not allowed"
            logger.error(msg)
            raise BackendParameterException(msg)

        data = iter(data)
        try:
            first_record = next(data)
        except StopIteration:
            logger.info("Data Iterator is empty; skipping write to target")
            return 0
        data = chain((first_record,), data)

        chunk_size = chunk_size if chunk_size else self.settings.WRITE_CHUNK_SIZE
        is_bytes = isinstance(first_record, bytes)
        writer = self._write_bytes if is_bytes else self._write_dicts

        concurrency = concurrency if concurrency else 1
        if concurrency == 1:
            return await writer(data, target, chunk_size, ignore_errors, operation_type)

        if concurrency < 1:
            msg = "concurrency must be a strictly positive integer"
            logger.error(msg)
            raise BackendParameterException(msg)

        count = 0
        for batch in iter_by_batch(iter_by_batch(data, chunk_size), concurrency):
            tasks = set()
            for chunk in batch:
                task = writer(chunk, target, chunk_size, ignore_errors, operation_type)
                tasks.add(task)
            result = await gather_with_limited_concurrency(concurrency, *tasks)
            count += sum(result)
        return count

    async def _write_bytes(  # noqa: PLR0913
        self,
        data: Iterable[bytes],
        target: Optional[str],
        chunk_size: int,
        ignore_errors: bool,
        operation_type: BaseOperationType,
    ) -> int:
        """Method called by `self.write` writing bytes. See `self.write`."""
        statements = parse_iterable_to_dict(data, ignore_errors)
        return await self._write_dicts(
            statements, target, chunk_size, ignore_errors, operation_type
        )

    @abstractmethod
    async def _write_dicts(  # noqa: PLR0913
        self,
        data: Iterable[dict],
        target: Optional[str],
        chunk_size: int,
        ignore_errors: bool,
        operation_type: BaseOperationType,
    ) -> int:
        """Method called by `self.write` writing dictionaries. See `self.write`."""
        locale = self.settings.LOCALE_ENCODING
        statements = parse_dict_to_bytes(data, locale, ignore_errors)
        return await self._write_bytes(
            statements, target, chunk_size, ignore_errors, operation_type
        )


class AsyncListable(ABC):
    """Async data backend interface for backends supporting the list operation."""

    @abstractmethod
    async def list(
        self, target: Optional[str] = None, details: bool = False, new: bool = False
    ) -> Union[AsyncIterator[str], AsyncIterator[dict]]:
        """List containers in the data backend. E.g., collections, files, indexes.

        Args:
            target (str or None): The target container name.
                If `target` is `None`, a default value is used instead.
            details (bool): Get detailed container information instead of just names.
            new (bool): Given the history, list only not already read containers.

        Yield:
            str: If `details` is False.
            dict: If `details` is True.

        Raise:
            BackendException: If a failure occurs.
            BackendParameterException: If a backend argument value is not valid.
        """


class BaseAsyncDataBackend(Generic[Settings, Query], ABC):
    """Base async data backend interface."""

    name = "base"
    query_class: Type[Query]
    settings_class: Type[Settings]

    def __init_subclass__(cls, **kwargs: Any) -> None:  # noqa: D105
        super().__init_subclass__(**kwargs)
        set_backend_settings_class(cls)
        set_backend_query_class(cls)

    def __init__(self, settings: Optional[Settings] = None):
        """Instantiate the data backend.

        Args:
            settings (Settings or None): The backend settings.
                If `settings` is `None`, a default settings instance is used instead.
        """
        try:
            self.settings: Settings = settings if settings else self.settings_class()
        except ValidationError as error:
            msg = "Failed to instantiate default async data backend settings: %s"
            logger.error(msg, error)
            raise BackendParameterException(msg % error) from error

    @abstractmethod
    async def status(self) -> DataBackendStatus:
        """Implement data backend checks (e.g. connection, cluster status).

        Return:
            DataBackendStatus: The status of the data backend.
        """

    async def read(  # noqa: PLR0913
        self,
        query: Optional[Query] = None,
        target: Optional[str] = None,
        chunk_size: Optional[int] = None,
        raw_output: bool = False,
        ignore_errors: bool = False,
        prefetch: Optional[PositiveInt] = None,
        max_statements: Optional[PositiveInt] = None,
    ) -> Union[AsyncIterator[bytes], AsyncIterator[dict]]:
        """Read records matching the `query` in the `target` container and yield them.

        Args:
            query (Query): The query to select records to read.
            target (str or None): The target container name.
                If `target` is `None`, a default value is used instead.
            chunk_size (int or None): The number of records or bytes to read in one
                batch, depending on whether the records are dictionaries or bytes.
                If `chunk_size` is `None` it defaults to `READ_CHUNK_SIZE`.
            raw_output (bool): Controls whether to yield bytes or dictionaries.
                If the records are dictionaries and `raw_output` is set to `True`, they
                are encoded as JSON.
                If the records are bytes and `raw_output` is set to `False`, they are
                decoded as JSON by line.
            ignore_errors (bool): If `True`, encoding errors during the read operation
                will be ignored and logged.
                If `False` (default), a `BackendException` is raised on any error.
            prefetch (int): The number of records to prefetch (queue) while yielding.
                If `prefetch` is `None` it defaults to `1`, i.e. no records are
                prefetched.
            max_statements (int): The maximum number of statements to yield.
                If `None` (default) or `0`, there is no maximum.

        Yield:
            dict: If `raw_output` is False.
            bytes: If `raw_output` is True.

        Raise:
            BackendException: If a failure during the read operation occurs or
                during encoding records and `ignore_errors` is set to `False`.
            BackendParameterException: If a backend argument value is not valid.
        """
        prefetch = prefetch if prefetch else 1
        if prefetch < 1:
            msg = "prefetch must be a strictly positive integer"
            logger.error(msg)
            raise BackendParameterException(msg)

        if prefetch > 1:
            queue = Queue(prefetch - 1)
            statements = self.read(
                query,
                target,
                chunk_size,
                raw_output,
                ignore_errors,
                None,
                max_statements,
            )
            task = create_task(self._queue_records(queue, statements))
            while True:
                statement = await queue.get()
                if statement is None:
                    error = task.exception()
                    if error:
                        raise error

                    return

                yield statement

        chunk_size = chunk_size if chunk_size else self.settings.READ_CHUNK_SIZE
        query = validate_backend_query(query, self.query_class)
        reader = self._read_bytes if raw_output else self._read_dicts
        statements = reader(query, target, chunk_size, ignore_errors)
        if not max_statements:
            async for statement in statements:
                yield statement
            return

        i = 0
        async for statement in statements:
            yield statement
            i += 1
            if i >= max_statements:
                return

    async def _read_bytes(
        self, query: Query, target: Optional[str], chunk_size: int, ignore_errors: bool
    ) -> AsyncIterator[bytes]:
        """Method called by `self.read` yielding bytes. See `self.read`."""
        statements = self._read_dicts(query, target, chunk_size, ignore_errors)
        async for statement in async_parse_dict_to_bytes(
            statements, self.settings.LOCALE_ENCODING, ignore_errors
        ):
            yield statement

    @abstractmethod
    async def _read_dicts(
        self, query: Query, target: Optional[str], chunk_size: int, ignore_errors: bool
    ) -> AsyncIterator[dict]:
        """Method called by `self.read` yielding dictionaries. See `self.read`."""
        statements = self._read_bytes(query, target, chunk_size, ignore_errors)
        statements = async_parse_iterable_to_dict(statements, ignore_errors)
        async for statement in statements:
            yield statement

    @abstractmethod
    async def close(self) -> None:
        """Close the data backend client.

        Raise:
            BackendException: If a failure occurs during the close operation.
        """

    async def _queue_records(
        self, queue: Queue, records: Union[AsyncIterator[bytes], AsyncIterator[dict]]
    ):
        """Iterate over the `records` and put them into the `queue`."""
        try:
            async for record in records:
                await queue.put(record)
        except Exception as error:
            # None signals that the queue is done
            await queue.put(None)
            raise error

        await queue.put(None)


def get_backend_generic_argument(
    backend_class: Type[Union[BaseDataBackend, BaseAsyncDataBackend]],
    position: DataBackendArgument,
) -> Optional[Type]:
    """Return the generic argument of `backend_class` at specified `position`."""
    for base in get_original_bases(backend_class):
        origin = get_origin(base)
        if not (
            origin
            and isclass(origin)
            and issubclass(origin, (BaseDataBackend, BaseAsyncDataBackend))
        ):
            continue

        args = get_args(base)
        if len(args) < abs(position) + 1:
            return None

        argument = args[position]
        if argument is Any:
            return None

        if isinstance(argument, TypeVar):
            argument = argument.__bound__

        if get_origin(argument) is Union:
            argument = get_args(argument)[0]

        if isinstance(argument, Type):
            return argument

    return None


def set_backend_settings_class(
    backend_class: Type[Union[BaseDataBackend, BaseAsyncDataBackend]]
) -> None:
    """Set `settings_class` attribute with `Config.env_prefix` for `backend_class`."""
    settings_class = get_backend_generic_argument(
        backend_class, DataBackendArgument.SETTINGS
    )
    if settings_class:
        backend_class.settings_class = settings_class


def set_backend_query_class(
    backend_class: Type[Union[BaseDataBackend, BaseAsyncDataBackend]]
) -> None:
    """Set `query_class` attribute for `backend_class`."""
    query_class = get_backend_generic_argument(backend_class, DataBackendArgument.QUERY)
    if query_class:
        backend_class.query_class = query_class



"""`Scorm Profile` verbs definitions."""

import sys
from typing import Dict, Optional

from ...base.verbs import BaseXapiVerb
from ...constants import LANG_EN_US_DISPLAY

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class CompletedVerb(BaseXapiVerb):
    """Pydantic model for completed `verb`.

    Attributes:
        id (str): Consists of the value `http://adlnet.gov/expapi/verbs/completed`.
        display (dict): Consists of the dictionary `{"en-US": "completed"}`.
    """

    id: Literal[
        "http://adlnet.gov/expapi/verbs/completed"
    ] = "http://adlnet.gov/expapi/verbs/completed"
    display: Optional[Dict[Literal[LANG_EN_US_DISPLAY], Literal["completed"]]]


class InitializedVerb(BaseXapiVerb):
    """Pydantic model for initialized `verb`.

    Attributes:
        id (str): Consists of the value `http://adlnet.gov/expapi/verbs/initialized`.
        display (Dict): Consists of the dictionary `{"en-US": "initialized"}`.
    """

    id: Literal[
        "http://adlnet.gov/expapi/verbs/initialized"
    ] = "http://adlnet.gov/expapi/verbs/initialized"
    display: Optional[Dict[Literal[LANG_EN_US_DISPLAY], Literal["initialized"]]]


class InteractedVerb(BaseXapiVerb):
    """Pydantic model for interacted `verb`.

    Attributes:
        id (str): Consists of the value `http://adlnet.gov/expapi/verbs/interacted`.
        display (dict): Consists of the dictionary `{"en-US": "interacted"}`.
    """

    id: Literal[
        "http://adlnet.gov/expapi/verbs/interacted"
    ] = "http://adlnet.gov/expapi/verbs/interacted"
    display: Optional[Dict[Literal[LANG_EN_US_DISPLAY], Literal["interacted"]]]


class TerminatedVerb(BaseXapiVerb):
    """Pydantic model for terminated `verb`.

    Attributes:
        id (str): Consists of the value `http://adlnet.gov/expapi/verbs/terminated`.
        display (dict): Consists of the dictionary `{"en-US": "terminated"}`.
    """

    id: Literal[
        "http://adlnet.gov/expapi/verbs/terminated"
    ] = "http://adlnet.gov/expapi/verbs/terminated"
    display: Optional[Dict[Literal[LANG_EN_US_DISPLAY], Literal["terminated"]]]



"""Tests for Ralph cli."""

import json
import logging
from contextlib import contextmanager
from importlib import reload
from pathlib import Path
from typing import Optional, Union

import pytest
from click.exceptions import BadParameter
from click.testing import CliRunner
from elasticsearch.helpers import bulk, scan
from hypothesis import settings as hypothesis_settings
from pydantic import ValidationError

from ralph import cli as cli_module
from ralph.backends.data.fs import FSDataBackend
from ralph.backends.data.ldp import LDPDataBackend
from ralph.cli import (
    CommaSeparatedKeyValueParamType,
    CommaSeparatedTupleParamType,
    JSONStringParamType,
    backends_options,
    cli,
)
from ralph.conf import settings
from ralph.exceptions import BackendParameterException, ConfigurationException
from ralph.models.edx.navigational.statements import UIPageClose
from ralph.models.xapi.navigation.statements import PageTerminated
from ralph.utils import iter_over_async

from tests.fixtures.backends import (
    ES_TEST_HOSTS,
    ES_TEST_INDEX,
    WS_TEST_HOST,
    WS_TEST_PORT,
)
from tests.fixtures.hypothesis_strategies import custom_given

test_logger = logging.getLogger("ralph")


def test_cli_comma_separated_key_value_param_type():
    """Test the CommaSeparatedKeyValueParamType custom parameter type."""
    param_type = CommaSeparatedKeyValueParamType()

    # Bad options
    with pytest.raises(
        BadParameter,
        match=(
            "You should provide key=value pairs separated by commas, "
            "e.g. foo=bar,bar=2"
        ),
    ):
        param_type.convert("foo=bar,baz", None, None)
    with pytest.raises(
        BadParameter,
        match=(
            "You should provide key=value pairs separated by commas, "
            "e.g. foo=bar,bar=2"
        ),
    ):
        param_type.convert("foo=bar,", None, None)

    # Option already a dictionary (from defaults)
    assert param_type.convert({"foo": "bar"}, None, None) == {"foo": "bar"}

    # Simple string parsing
    assert param_type.convert("foo=bar", None, None) == {
        "foo": "bar",
    }
    assert param_type.convert("foo=bar,baz=spam", None, None) == {
        "foo": "bar",
        "baz": "spam",
    }
    assert param_type.convert("foo=bar,baz=", None, None) == {
        "foo": "bar",
        "baz": None,
    }

    # Boolean type casting
    assert param_type.convert("foo=true,bar=false", None, None) == {
        "foo": True,
        "bar": False,
    }
    assert param_type.convert("foo=True,bar=False", None, None) == {
        "foo": True,
        "bar": False,
    }

    # Integer type casting
    assert param_type.convert("foo=42,bar=1", None, None) == {
        "foo": 42,
        "bar": 1,
    }

    # Float type casting
    assert param_type.convert("foo=4.2,bar=12.3", None, None) == {
        "foo": 4.2,
        "bar": 12.3,
    }

    # Mixed types casting
    assert param_type.convert("foo=lol,bar=12.3,baz=1,spam=True", None, None) == {
        "foo": "lol",
        "bar": 12.3,
        "baz": 1,
        "spam": True,
    }


@pytest.mark.parametrize(
    "value,expected",
    [
        ("foo", ("foo",)),
        ("foo,bar", ("foo", "bar")),
        ("foo,bar,baz", ("foo", "bar", "baz")),
        (("foo", "bar", "baz"), ("foo", "bar", "baz")),
    ],
)
def test_cli_comma_separated_tuple_param_type_with_valid_input(value, expected):
    """Test the CommaSeparatedTupleParamType parameter type with valid input."""
    param_type = CommaSeparatedTupleParamType()
    assert param_type.convert(value, None, None) == expected


@pytest.mark.parametrize("value", [None, {}, ["foo"], 10, True])
def test_cli_comma_separated_tuple_param_type_with_invalid_input(value):
    """Test the CommaSeparatedTupleParamType parameter type with invalid input."""
    param_type = CommaSeparatedTupleParamType()
    with pytest.raises(
        BadParameter,
        match="You should provide values separated by commas, e.g. foo,bar,baz",
    ):
        param_type.convert(value, None, None)


@pytest.mark.parametrize("value,expected", [('{"foo": "bar"}', {"foo": "bar"})])
def test_cli_json_string_param_type_with_valid_input(value, expected):
    """Test the JSONStringParamType custom parameter type with valid input."""
    param_type = JSONStringParamType()
    assert param_type.convert(value, None, None) == expected


@pytest.mark.parametrize("value", ["foo", None, {}])
def test_cli_json_string_param_type_with_invalid_input(value):
    """Test the JSONStringParamType custom parameter type with invalid input."""
    param_type = JSONStringParamType()
    with pytest.raises(
        BadParameter,
        match="You should provide a valid JSON string as input",
    ):
        param_type.convert(value, None, None)


def test_cli_help_option():
    """Test ralph --help command."""
    runner = CliRunner()
    result = runner.invoke(cli, ["--help"])

    assert result.exit_code == 0
    assert (
        (
            "-v, --verbosity LVL  Either CRITICAL, ERROR, WARNING, INFO (default) or "
            "DEBUG"
        )
    ) in result.output


def _gen_cli_auth_args(  # noqa: PLR0913
    username: str,
    password: str,
    scopes: list,
    ifi_command: str,
    ifi_value: Union[str, dict],
    agent_name: Optional[str] = None,
    target: Optional[str] = None,
    write: bool = False,
):
    """Generate arguments for cli to create user."""
    cli_args = ["auth", "-u", username, "-p", password]
    for scope in scopes:
        cli_args.extend(["-s", scope])
    if target:
        cli_args.extend(["-t", target])
    cli_args.append(ifi_command)
    cli_args.extend(ifi_value.split())
    if agent_name:
        cli_args.extend(["-N", agent_name])
    if write:
        cli_args.append("-w")
    return cli_args


def _assert_matching_basic_auth_credentials(  # noqa: PLR0913
    credentials: dict,
    username: str,
    scopes: list,
    ifi_type: str,
    ifi_value: Union[str, dict],
    agent_name: str = None,
    hash_: str = None,
    target: str = None,
):
    """Assert that credentials match other arguments.

    Args:
        credentials: credentials to match against
        username: username that should match credentials
        scopes: scopes that should match credentials
        ifi_type: type of ifi used in credentials. One of "mbox",
            "mbox_sha1sum", "openid", or "account"
        ifi_value: value of ifi
        agent_name: name to assign to agent. If None, no matching is performed
        hash: hash value of the password. If None, no matching is performed
        target: target that should match credentials. If None, no matching is performed
    """
    assert credentials["username"] == username
    assert "hash" in credentials
    if hash_:
        assert credentials["hash"] == hash_
    assert sorted(credentials["scopes"]) == sorted(scopes)
    assert "agent" in credentials

    if target is not None:
        assert credentials["target"] == target

    if agent_name is not None:
        assert credentials["agent"]["name"] == agent_name

    if "objectType" in credentials["agent"]:
        assert credentials["agent"]["objectType"] == "Agent"

    if ifi_type in ["mbox", "mbox_sha1sum", "openid"]:
        assert ifi_type in credentials["agent"]
        assert credentials["agent"][ifi_type] == ifi_value
    elif ifi_type == "account":
        assert "account" in credentials["agent"]
        assert "name" in credentials["agent"]["account"]
        assert "homePage" in credentials["agent"]["account"]
        assert credentials["agent"]["account"] == ifi_value
    else:
        raise ValueError(
            'ifi_type should be one of: "mbox", "mbox_sha1sum", "openid", "account"'
        )


def _ifi_type_from_command(ifi_command):
    """Return the ifi_type associated to the command being passed to cli."""
    if ifi_command not in ["-M", "-S", "-O", "-A"]:
        raise ValueError('The ifi_command must be one of: "-M", "-S", "-O", "-A"')

    return {"-M": "mbox", "-S": "mbox_sha1sum", "-O": "openid", "-A": "account"}[
        ifi_command
    ]


def _ifi_value_from_command(ifi_value, ifi_type):
    """Parse ifi_value returned by cli to generate dict when `ifi_type` is `account`."""
    if ifi_type == "account":
        # Parse arguments from cli
        return {"name": ifi_value.split()[0], "homePage": ifi_value.split()[1]}
    return ifi_value


@pytest.mark.parametrize(
    "scopes,ifi_command,ifi_value",
    [
        (["all"], "-M", "mailto:foo@bar.com"),  # mbox ifi
        (["all/read", "all"], "-M", "mailto:foo@bar.com"),
        (
            ["all"],
            "-S",
            "ebd31e95054c018b10727ccffd2ef2ec3a016ee9",
        ),  # mbox_sha1sum ifi
        (["all"], "-O", "http://foo.openid.example.org/"),  # mbox_openid ifi
        (["all"], "-A", "foo_name http://www.bar.xyz"),  # account ifi
    ],
)
def test_cli_auth_command_without_writing_auth_file(scopes, ifi_command, ifi_value):
    """Test ralph auth command when credentials are displayed in the tty."""

    username = "foo"
    password = "bar"
    agent_name = "foobarname"  # optional name
    target = "custom_target"

    cli_args = _gen_cli_auth_args(
        username,
        password,
        scopes,
        ifi_command,
        ifi_value,
        agent_name=agent_name,
        target=target,
        write=False,
    )

    runner = CliRunner()
    result = runner.invoke(cli, cli_args)

    assert result.exit_code == 0

    credentials = json.loads(result.output.split(".json", 1)[1].strip())

    ifi_type = _ifi_type_from_command(ifi_command=ifi_command)
    ifi_value = _ifi_value_from_command(ifi_value, ifi_type)

    _assert_matching_basic_auth_credentials(
        credentials=credentials,
        username=username,
        scopes=scopes,
        ifi_type=ifi_type,
        ifi_value=ifi_value,
        agent_name=agent_name,
        target=target,
    )


@pytest.mark.parametrize(
    "ifi_command_1, ifi_value_1, ifi_command_2, ifi_value_2",
    [
        (
            "-M",
            "mailto:foo@bar.com",
            "-S",
            "ebd31e95054c018b10727ccffd2ef2ec3a016ee9",
        ),  # mbox ifi
        (
            "-S",
            "ebd31e95054c018b10727ccffd2ef2ec3a016ee9",
            "-O",
            "http://foo.openid.example.org/",
        ),  # mbox_sha1sum ifi
        (
            "-O",
            "http://foo.openid.example.org/",
            "-A",
            "foo_name http://www.bar.xyz",
        ),  # mbox_openid ifi
        (
            "-A",
            "foo_name http://www.bar.xyz",
            "-M",
            "mailto:foo@bar.com",
        ),  # account ifi
    ],
)
def test_cli_auth_command_when_writing_auth_file(
    fs, ifi_command_1, ifi_value_1, ifi_command_2, ifi_value_2
):
    """Test ralph auth command when credentials are written in the authentication
    file.
    """
    runner = CliRunner()

    username_1 = "foo"
    password_1 = "bar"
    scopes_1 = ["all"]
    target_1 = "custom_target"

    # The authentication file does not exist

    # Add a first user
    cli_args = _gen_cli_auth_args(
        username_1,
        password_1,
        scopes_1,
        ifi_command_1,
        ifi_value_1,
        target=target_1,
        write=True,
    )

    assert Path(settings.AUTH_FILE).exists() is False
    result = runner.invoke(cli, cli_args)
    assert result.exit_code == 0
    assert Path(settings.AUTH_FILE).exists() is True
    with Path(settings.AUTH_FILE).open(encoding="utf-8") as auth_file:
        all_credentials = json.loads("\n".join(auth_file.readlines()))
    assert len(all_credentials) == 1

    # Check that the first user matches
    ifi_type_1 = _ifi_type_from_command(ifi_command=ifi_command_1)
    ifi_value_1 = _ifi_value_from_command(ifi_value_1, ifi_type_1)
    _assert_matching_basic_auth_credentials(
        credentials=all_credentials[0],
        username=username_1,
        scopes=scopes_1,
        ifi_type=ifi_type_1,
        ifi_value=ifi_value_1,
        target=target_1,
    )

    # Add a second user
    username_2 = "lol"
    password_2 = "baz"
    scopes_2 = ["statements/write", "statements/read/mine"]
    target_2 = "there"

    cli_args = _gen_cli_auth_args(
        username_2,
        password_2,
        scopes_2,
        ifi_command_2,
        ifi_value_2,
        target=target_2,
        write=True,
    )
    result = runner.invoke(cli, cli_args)

    assert result.exit_code == 0
    with Path(settings.AUTH_FILE).open(encoding="utf-8") as auth_file:
        all_credentials = json.loads("\n".join(auth_file.readlines()))
    assert len(all_credentials) == 2

    # Check that the first user still matches
    _assert_matching_basic_auth_credentials(
        credentials=all_credentials[0],
        username=username_1,
        scopes=scopes_1,
        ifi_type=ifi_type_1,
        ifi_value=ifi_value_1,
        target=target_1,
    )

    # Check that the second user matches
    ifi_type_2 = _ifi_type_from_command(ifi_command=ifi_command_2)
    ifi_value_2 = _ifi_value_from_command(ifi_value_2, ifi_type_2)
    _assert_matching_basic_auth_credentials(
        credentials=all_credentials[1],
        username=username_2,
        scopes=scopes_2,
        ifi_type=ifi_type_2,
        ifi_value=ifi_value_2,
        target=target_2,
    )


def test_cli_auth_command_when_writing_auth_file_with_incorrect_auth_file(fs):
    """Test ralph auth command when credentials are written in the authentication
    file with a badly formatted original authentication file.
    """
    runner = CliRunner()

    # Create the base auth file path in the fake fs with badly formatted JSON
    # file content
    contents = (
        "{\n"
        ' "username": "johndoe",\n'
        ' "hash": "secretlynothashed",\n'
        ' "scope": ["johndoe_scope"]\n'
    )
    fs.create_file(settings.AUTH_FILE, contents=contents)

    result = runner.invoke(
        cli, ["auth", "-u", "foo", "-p", "bar", "-s", "foo_scope", "-w"]
    )
    assert result.exit_code > 0
    auth_file = Path(settings.AUTH_FILE)
    assert auth_file.exists() is True
    assert auth_file.read_text(encoding="utf-8") == contents


def test_cli_extract_command_with_gelf_parser(gelf_logger):
    """Test ralph extract command using the GELF parser."""
    gelf_logger.info('{"username": "foo"}')

    runner = CliRunner()
    with Path(gelf_logger.handlers[0].stream.name).open(
        encoding=settings.LOCALE_ENCODING
    ) as log_file:
        gelf_content = log_file.read()
        result = runner.invoke(cli, ["extract", "-p", "gelf"], input=gelf_content)
        assert '{"username": "foo"}' in result.output


def test_cli_extract_command_with_es_parser():
    """Test ralph extract command using the ElasticSearchParser."""
    es_output = (
        "\n".join(
            [
                json.dumps(
                    {
                        "_index": ES_TEST_INDEX,
                        "_id": str(idx),
                        "_score": None,
                        "_source": {"id": idx},
                        "sort": [idx],
                    }
                )
                for idx in range(10)
            ]
        )
        + "\n"
    )

    runner = CliRunner()
    result = runner.invoke(cli, "extract -p es".split(), input=es_output)
    assert result.exit_code == 0
    assert "\n".join([json.dumps({"id": idx}) for idx in range(10)]) in result.output


@custom_given(UIPageClose)
def test_cli_validate_command_with_edx_format(event):
    """Test ralph validate command using the edx format."""
    event_str = event.json()
    runner = CliRunner()
    result = runner.invoke(cli, ["validate", "-f", "edx"], input=event_str)
    assert event_str in result.output


@hypothesis_settings(deadline=None)
@custom_given(UIPageClose)
@pytest.mark.parametrize("valid_uuid", ["ee241f8b-174f-5bdb-bae9-c09de5fe017f"])
def test_cli_convert_command_from_edx_to_xapi_format(valid_uuid, event):
    """Test ralph convert command from edx to xapi format."""
    event_str = event.json()
    runner = CliRunner()
    command = f"-v ERROR convert -f edx -t xapi -u {valid_uuid} -p https://fun-mooc.fr"
    result = runner.invoke(cli, command.split(), input=event_str)
    assert result.exit_code == 0
    try:
        PageTerminated(**json.loads(result.output))
    except ValidationError as err:
        pytest.fail(f"Converted event is invalid: {err}")


@pytest.mark.parametrize("invalid_uuid", ["", None, 1, {}])
def test_cli_convert_command_with_invalid_uuid(invalid_uuid):
    """Test that the ralph convert command raises an exception when the uuid namespace
    is invalid.
    """
    runner = CliRunner()
    command = f"convert -f edx -t xapi -u '{invalid_uuid}' -p https://fun-mooc.fr"
    result = runner.invoke(cli, command.split())
    assert result.exit_code > 0
    assert isinstance(result.exception, ConfigurationException)
    assert str(result.exception) == "Invalid UUID namespace"


@cli.command()
def dummy_verbosity_check():
    """Adding a dummy command to the cli with all logging levels."""
    test_logger.critical("CRITICAL")
    test_logger.error("ERROR")
    test_logger.warning("WARNING")
    test_logger.info("INFO")
    test_logger.debug("DEBUG")


@pytest.mark.parametrize("verbosity", ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"])
def test_cli_verbosity_option_should_impact_logging_behaviour(verbosity):
    """Test that the verbosity option impacts logging output."""
    runner = CliRunner()
    result = runner.invoke(cli, ["-v", verbosity, "dummy-verbosity-check"])
    assert verbosity in result.output


def test_cli_read_command_with_ldp_backend(monkeypatch):
    """Test ralph read command using the LDP backend."""
    archive_content = {"foo": "bar"}

    def mock_read(*_, **__):
        """Always return the same archive."""

        yield bytes(json.dumps(archive_content), encoding="utf-8")

    monkeypatch.setattr(LDPDataBackend, "read", mock_read)

    runner = CliRunner()
    command = "read -b ldp --ldp-endpoint ovh-eu a547d9b3-6f2f-4913-a872-cf4efe699a66"
    result = runner.invoke(cli, command.split())

    assert result.exit_code == 0
    assert '{"foo": "bar"}' in result.output


def test_cli_read_command_with_fs_backend(fs, monkeypatch):
    """Test ralph read command using the FS backend."""
    archive_content = {"foo": "bar"}

    def mock_read(*_, **__):
        """Always return the same archive."""

        yield bytes(json.dumps(archive_content), encoding="utf-8")

    monkeypatch.setattr(FSDataBackend, "read", mock_read)

    runner = CliRunner()
    result = runner.invoke(cli, "read -b fs foo".split())

    assert result.exit_code == 0
    assert '{"foo": "bar"}' in result.output


def test_cli_read_command_with_chunk_size(fs, monkeypatch):
    """Test ralph `read` command with a `chunk_size` option."""

    def get_mock_read_bytes(expected_chunk_size: int):
        """Return a `_read_bytes` function checking the `chunk_size` option."""

        def mock_read_bytes(self, query, target, chunk_size, *_):
            """Check `chunk_size` and return an archive."""
            assert chunk_size == expected_chunk_size
            data = b'{"foo": "bar"}'
            for i in range(0, len(data), chunk_size):
                yield data[i : i + chunk_size]

        return mock_read_bytes

    monkeypatch.delenv("RALPH_BACKENDS__DATA__FS__READ_CHUNK_SIZE", raising=False)
    runner = CliRunner()

    # Given no chunk size, a default chunk size should be used.
    monkeypatch.setattr(FSDataBackend, "_read_bytes", get_mock_read_bytes(4096))
    result = runner.invoke(cli, "read -b fs".split())
    assert result.exit_code == 0
    assert '{"foo": "bar"}' in result.output

    # Given a chunk size set by the environment, it should overwrite the default.
    runner = CliRunner(env={"RALPH_BACKENDS__DATA__FS__READ_CHUNK_SIZE": "3"})
    monkeypatch.setattr(FSDataBackend, "_read_bytes", get_mock_read_bytes(3))
    result = runner.invoke(cli, "read -b fs".split())
    assert result.exit_code == 0
    assert '{"foo": "bar"}' in result.output

    # Given a chunk size set by the chunk-size option, it should overwrite the default.
    monkeypatch.setattr(FSDataBackend, "_read_bytes", get_mock_read_bytes(1))
    result = runner.invoke(cli, "read -b fs --chunk-size 1".split())
    assert result.exit_code == 0
    assert '{"foo": "bar"}' in result.output


def test_cli_read_command_with_es_backend(es):
    """Test ralph `read` command using the es backend."""

    # Insert documents
    bulk(
        es,
        (
            {"_index": ES_TEST_INDEX, "_id": idx, "_source": {"id": idx}}
            for idx in range(10)
        ),
    )
    # As we bulk insert documents, the index needs to be refreshed before making
    # queries.
    es.indices.refresh(index=ES_TEST_INDEX)

    runner = CliRunner()
    es_hosts = ",".join(ES_TEST_HOSTS)
    es_client_options = "verify_certs=True"
    command = f"""-v ERROR read -b es --es-hosts {es_hosts}
        --es-default-index {ES_TEST_INDEX}
        --es-client-options {es_client_options}"""
    result = runner.invoke(cli, command.split())
    assert result.exit_code == 0
    expected = (
        "\n".join(
            [
                json.dumps(
                    {
                        "_index": ES_TEST_INDEX,
                        "_id": str(idx),
                        "_score": None,
                        "_source": {"id": idx},
                        "sort": [idx],
                    }
                )
                for idx in range(10)
            ]
        )
        + "\n"
    )

    assert expected == result.output


def test_cli_read_command_client_options_with_es_backend(es):
    """Test ralph read command with multiple client options using the es backend."""

    runner = CliRunner()
    es_client_options = "ca_certs=/path/,verify_certs=True"
    command = f"""-v ERROR read -b es --es-client-options {es_client_options}"""
    result = runner.invoke(cli, command.split())
    assert result.exit_code == 1
    assert "TLS options require scheme to be 'https'" in str(result.exception)


def test_cli_read_command_with_es_backend_query(es):
    """Test ralph read command using the es backend and a query."""

    # Insert documents
    bulk(
        es,
        (
            {
                "_index": ES_TEST_INDEX,
                "_id": idx,
                "_source": {"id": idx, "modulo": idx % 2},
            }
            for idx in range(10)
        ),
    )
    # As we bulk insert documents, the index needs to be refreshed before making
    # queries.
    es.indices.refresh(index=ES_TEST_INDEX)

    runner = CliRunner()
    es_hosts = ",".join(ES_TEST_HOSTS)
    query = {"query": {"term": {"modulo": 0}}}
    query_str = json.dumps(query, separators=(",", ":"))

    command = (
        "-v ERROR "
        "read "
        "-b es "
        f"--es-hosts {es_hosts} "
        f"--es-default-index {ES_TEST_INDEX} "
        f"{query_str}"
    )
    result = runner.invoke(cli, command.split())
    assert result.exit_code == 0
    expected = (
        "\n".join(
            [
                json.dumps(
                    {
                        "_index": ES_TEST_INDEX,
                        "_id": str(idx),
                        "_score": None,
                        "_source": {"id": idx, "modulo": 0},
                        "sort": [idx],
                    }
                )
                for idx in range(0, 10, 2)
            ]
        )
        + "\n"
    )

    assert expected == result.output

    # Test with an invalid json query string
    invalid_query_str = "wrong_query_string"

    command = f"-v ERROR read -b mongo {invalid_query_str}"
    result = runner.invoke(cli, command.split())
    assert result.exit_code > 0
    assert isinstance(result.exception, BackendParameterException)
    msg = (
        "Invalid MongoQuery query string: "
        "[{'loc': ('__root__',), 'msg': 'Expecting value: line 1 column 1 (char 0)', "
        "'type': 'value_error.jsondecode', 'ctx': {'msg': 'Expecting value', "
        "'doc': 'wrong_query_string', 'pos': 0, 'lineno': 1, 'colno': 1}}]"
    )
    assert str(result.exception) == msg


def test_cli_read_command_with_ws_backend(events, ws):
    """Test ralph read command using the ws backend."""

    # The ws fixture is async, however the `CliRunner` does not support running in an
    # async context, thus we wrap it into a sync contextmanager.
    @contextmanager
    def websocket():
        yield from iter_over_async(ws)

    with websocket():
        runner = CliRunner()
        uri = f"ws://{WS_TEST_HOST}:{WS_TEST_PORT}"
        result = runner.invoke(cli, ["read", "-b", "async_ws", "--async-ws-uri", uri])
        assert "\n".join([json.dumps(event) for event in events]) in result.output


def test_cli_list_command_with_ldp_backend(monkeypatch):
    """Test ralph list command using the LDP backend."""
    archive_list = [
        "5d5c4c93-04a4-42c5-9860-f51fa4044aa1",
        "997db3eb-b9ca-485d-810f-b530a6cef7c6",
    ]
    archive_list_details = [
        {
            "archiveId": "5d5c4c93-04a4-42c5-9860-f51fa4044aa1",
            "createdAt": "2020-06-18T04:38:59.436634+02:00",
            "filename": "2020-06-16.gz",
            "md5": "01585b394be0495e38dbb60b20cb40a9",
            "retrievalDelay": 0,
            "retrievalState": "sealed",
            "sha256": "645d8e21e6fdb8aa7ffc5c[...]9ce612d06df8dcf67cb29a45ca",
            "size": 67906662,
        },
        {
            "archiveId": "997db3eb-b9ca-485d-810f-b530a6cef7c6",
            "createdAt": "2020-06-18T04:38:59.436634+02:00",
            "filename": "2020-06-17.gz",
            "md5": "01585b394be0495e38dbb60b20cb40a9",
            "retrievalDelay": 0,
            "retrievalState": "sealed",
            "sha256": "645d8e21e6fdb8aa7ffc5c[...]9ce612d06df8dcf67cb29a45ca",
            "size": 67906662,
        },
    ]

    def mock_list(this, target=None, details=False, new=False):
        """Mock LDP backend list method."""

        response = archive_list
        if details:
            response = archive_list_details
        if new:
            response = response[1:]
        return response

    monkeypatch.setattr(LDPDataBackend, "list", mock_list)

    runner = CliRunner()

    # List documents with default options
    result = runner.invoke(cli, ["list", "-b", "ldp", "--ldp-endpoint", "ovh-eu"])
    assert result.exit_code == 0
    assert "\n".join(archive_list) in result.output

    # List documents with detailed output
    result = runner.invoke(cli, ["list", "-b", "ldp", "--ldp-endpoint", "ovh-eu", "-D"])
    assert result.exit_code == 0
    assert (
        "\n".join(json.dumps(detail) for detail in archive_list_details)
        in result.output
    )

    # List new documents only
    result = runner.invoke(cli, ["list", "-b", "ldp", "--ldp-endpoint", "ovh-eu", "-n"])
    assert result.exit_code == 0
    assert "997db3eb-b9ca-485d-810f-b530a6cef7c6" in result.output
    assert "5d5c4c93-04a4-42c5-9860-f51fa4044aa1" not in result.output

    # Edge case: stream contains no document
    monkeypatch.setattr(LDPDataBackend, "list", lambda this, target, details, new: ())
    result = runner.invoke(cli, ["list", "-b", "ldp", "--ldp-endpoint", "ovh-eu"])
    assert result.exit_code == 0
    assert "Configured ldp backend contains no document" in result.output


def test_cli_list_command_with_fs_backend(fs, monkeypatch):
    """Test ralph list command using the LDP backend."""
    archive_list = [
        "file1",
        "file2",
    ]
    archive_list_details = [
        {
            "filename": "file1",
            "modified_at": "1604661859",
            "size": 350,
        },
        {
            "filename": "file2",
            "created_at": "1604935848",
            "size": 25000,
        },
    ]

    def mock_list(this, target=None, details=False, new=False):
        """Mock LDP backend list method."""

        response = archive_list
        if details:
            response = archive_list_details
        if new:
            response = response[1:]
        return response

    monkeypatch.setattr(FSDataBackend, "list", mock_list)

    runner = CliRunner()

    # List documents with default options
    result = runner.invoke(cli, ["list", "-b", "fs"])
    assert result.exit_code == 0
    assert "\n".join(archive_list) in result.output

    # List documents with detailed output
    result = runner.invoke(cli, ["list", "-b", "fs", "-D"])
    assert result.exit_code == 0
    assert (
        "\n".join(json.dumps(detail) for detail in archive_list_details)
        in result.output
    )

    # List new documents only
    result = runner.invoke(cli, ["list", "-b", "fs", "-n"])
    assert result.exit_code == 0
    assert "file2" in result.output
    assert "file1" not in result.output

    # Edge case: stream contains no document
    monkeypatch.setattr(FSDataBackend, "list", lambda this, target, details, new: ())
    result = runner.invoke(cli, ["list", "-b", "fs"])
    assert result.exit_code == 0
    assert "Configured fs backend contains no document" in result.output


def test_cli_write_command_with_fs_backend(fs):
    """Test ralph write command using the FS backend."""
    fs.create_dir(str(settings.APP_DIR))
    fs.create_dir("foo")

    filename = Path("foo/file1")

    # Create a file
    runner = CliRunner()
    result = runner.invoke(
        cli,
        "write -b fs -t file1 --fs-default-directory-path foo".split(),
        input=b"test content",
    )

    assert result.exit_code == 0

    with filename.open("rb") as test_file:
        content = test_file.read()

    assert b"test content" in content

    # Trying to create the same file without -f should raise an error
    runner = CliRunner()
    result = runner.invoke(
        cli,
        "write -b fs -t file1 --fs-default-directory-path foo".split(),
        input=b"other content",
    )
    assert result.exit_code == 1
    assert "file1 already exists and overwrite is not allowed" in result.output

    # Try to create the same file with -o update
    runner = CliRunner()
    result = runner.invoke(
        cli,
        "write -b fs -t file1 -o update --fs-default-directory-path foo".split(),
        input=b"other content",
    )

    assert result.exit_code == 0

    with filename.open("rb") as test_file:
        content = test_file.read()

    assert b"other content" in content


def test_cli_write_command_with_chunk_size(fs, monkeypatch):
    """Test ralph `write` command with a `chunk_size` option."""

    def get_mock_write_bytes(expected_chunk_size: int):
        """Return a `_write_bytes` function checking the `chunk_size` option."""

        def mock_write_bytes(self, query, target, chunk_size, *_):
            """Check `chunk_size` and return an archive."""
            assert chunk_size == expected_chunk_size
            return 1

        return mock_write_bytes

    monkeypatch.delenv("RALPH_BACKENDS__DATA__FS__WRITE_CHUNK_SIZE", raising=False)
    runner = CliRunner()

    # Given no chunk size, a default chunk size should be used.
    monkeypatch.setattr(FSDataBackend, "_write_bytes", get_mock_write_bytes(4096))
    result = runner.invoke(cli, "write -b fs".split())
    assert result.exit_code == 0

    # Given a chunk size set by the environment, it should overwrite the default.
    runner = CliRunner(env={"RALPH_BACKENDS__DATA__FS__WRITE_CHUNK_SIZE": "3"})
    monkeypatch.setattr(FSDataBackend, "_write_bytes", get_mock_write_bytes(3))
    result = runner.invoke(cli, "write -b fs".split())
    assert result.exit_code == 0

    # Given a chunk size set by the chunk-size option, it should overwrite the default.
    monkeypatch.setattr(FSDataBackend, "_write_bytes", get_mock_write_bytes(1))
    result = runner.invoke(cli, "write -b fs --chunk-size 1".split())
    assert result.exit_code == 0


def test_cli_write_command_with_es_backend(es):
    """Test ralph write command using the es backend."""

    # Documents
    records = [{"id": idx} for idx in range(10)]

    runner = CliRunner()
    es_hosts = ",".join(ES_TEST_HOSTS)
    result = runner.invoke(
        cli,
        f"write -b es --es-hosts {es_hosts} --es-default-index {ES_TEST_INDEX}".split(),
        input="\n".join(json.dumps(record) for record in records),
    )
    assert result.exit_code == 0

    # As we bulk insert documents, the index needs to be refreshed before making
    # queries.
    es.indices.refresh(index=ES_TEST_INDEX)
    documents = list(scan(es, index=ES_TEST_INDEX, size=10))

    assert len(documents) == 10
    assert [document.get("_source") for document in documents] == records


@pytest.mark.parametrize("host_,port_", [("0.0.0.0", "8000"), ("127.0.0.1", "80")])
def test_cli_runserver_command_with_host_and_port_arguments(host_, port_, monkeypatch):
    """Test the ralph runserver command should consider the host and port arguments."""

    def mock_uvicorn_run(_, host=None, port=None, **kwargs):
        """Mock uvicorn.run asserting host and port values."""
        assert host == host_
        assert port == int(port_)

    monkeypatch.setattr("ralph.cli.uvicorn.run", mock_uvicorn_run)

    runner = CliRunner()
    result = runner.invoke(cli, f"runserver -h {host_} -p {port_} -b es".split())
    assert result.exit_code == 0
    assert f"Running API server on {host_}:{port_} with es backend" in result.output


def test_cli_runserver_command_environment_file_generation(monkeypatch):
    """Test the ralph runserver command should create the expected environment file."""

    def mock_uvicorn_run(_, env_file=None, **kwargs):
        """Mock uvicorn.run asserting environment file content."""
        expected_env_lines = [
            f"RALPH_RUNSERVER_BACKEND={settings.RUNSERVER_BACKEND}\n",
            "RALPH_BACKENDS__LRS__ES__DEFAULT_INDEX=foo\n",
            "RALPH_BACKENDS__LRS__ES__CLIENT_OPTIONS__verify_certs=True\n",
        ]
        with open(env_file, mode="r", encoding=settings.LOCALE_ENCODING) as file:
            assert file.readlines() == expected_env_lines

    monkeypatch.setattr("ralph.cli.uvicorn.run", mock_uvicorn_run)
    runner = CliRunner()
    result = runner.invoke(
        cli,
        "runserver -b es --es-default-index foo "
        "--es-client-options verify_certs=True".split(),
    )
    assert result.exit_code == 0


def test_cli_ralph_cli_lazy_backend_options(monkeypatch):
    """Test the `RalphCli.lazy_backend_options` function."""

    reload(cli_module)
    call_counter = {"count": 0}

    def mock_backends_options(backends, name=None):
        call_counter["count"] += 1
        return backends_options(backends, name)

    monkeypatch.setattr("ralph.cli.backends_options", mock_backends_options)
    runner = CliRunner()
    # Given a command that does not require backend options, the `backend_options`
    # function should not be called.
    runner.invoke(cli_module.cli, ["convert --help"])
    assert not call_counter["count"]
    # Given a command that requires backend options, the `backend_options` function
    # should be called once.
    runner.invoke(cli_module.cli, ["list", "--help"])
    assert call_counter["count"] == 1
    # Given a command that requires backend options of multiple commands, the
    # `backend_options` function should be called once for each command.
    runner.invoke(cli_module.cli, ["--help"])
    assert call_counter["count"] == 4  # list + (read, write, runserver)



"""Base xAPI Converter."""

from typing import Set
from uuid import UUID, uuid5

from ralph.exceptions import ConfigurationException
from ralph.models.converter import BaseConversionSet, ConversionItem


class BaseXapiConverter(BaseConversionSet):
    """Base xAPI Converter.

    WARNING: The converter may not include the following edX fields:
    - context.org_id: When `org_id` is an empty string.
    - context.course_id: When `course_id` is an empty string.

    WARNING: The converter should not include the following edX fields as they may
    contain sensitive data: `username`, `referer`, `event`, `event_source`, `ip`,
    `agent`, `accept_language:`, `context.course_user_tags`.
    """

    def __init__(self, uuid_namespace: str, platform_url: str):
        """Initialize BaseXapiConverter."""
        self.platform_url = platform_url
        try:
            self.uuid_namespace = UUID(uuid_namespace)
        except (TypeError, ValueError, AttributeError) as err:
            raise ConfigurationException("Invalid UUID namespace") from err
        super().__init__()

    def _get_conversion_items(self) -> Set[ConversionItem]:
        """Return a set of ConversionItems used for conversion."""
        return {
            ConversionItem(
                "id",
                None,
                lambda event_str: str(uuid5(self.uuid_namespace, event_str)),
                True,
            ),
            ConversionItem(
                "actor__account__homePage", transformers=lambda _: self.platform_url
            ),
            ConversionItem(
                "actor__account__name",
                "context__user_id",
                lambda user_id: str(user_id) if user_id else "anonymous",
            ),
            ConversionItem("timestamp", "time"),
        }



"""Authenticated user for the Ralph API."""

from typing import Dict, FrozenSet, Literal, Optional

from pydantic import BaseModel

Scope = Literal[
    "statements/write",
    "statements/read/mine",
    "statements/read",
    "state/write",
    "state/read",
    "define",
    "profile/write",
    "profile/read",
    "all/read",
    "all",
]


class UserScopes(FrozenSet[Scope]):
    """Scopes available to users."""

    def is_authorized(self, requested_scope: Scope):
        """Check if the requested scope can be accessed based on user scopes."""
        expanded_scopes = {
            "statements/read": {"statements/read/mine", "statements/read"},
            "all/read": {
                "statements/read/mine",
                "statements/read",
                "state/read",
                "profile/read",
                "all/read",
            },
            "all": {
                "statements/write",
                "statements/read/mine",
                "statements/read",
                "state/read",
                "state/write",
                "define",
                "profile/read",
                "profile/write",
                "all/read",
                "all",
            },
        }

        expanded_user_scopes = set()
        for scope in self:
            expanded_user_scopes.update(expanded_scopes.get(scope, {scope}))

        return requested_scope in expanded_user_scopes

    @classmethod
    def __get_validators__(cls):  # noqa: D105
        def validate(value: FrozenSet[Scope]):
            """Transform value to an instance of UserScopes."""
            return cls(value)

        yield validate


class AuthenticatedUser(BaseModel):
    """Pydantic model for user authentication.

    Attributes:
        agent (dict): The agent representing the current user.
        scopes (list): The scopes the user has access to.
        target (str or None): The target index or database to store statements into.
    """

    agent: Dict
    scopes: UserScopes
    target: Optional[str]



"""Enrollment models event field definition."""

import sys
from typing import Union

from ...base import AbstractBaseEventField

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class EnrollmentEventField(AbstractBaseEventField):
    """Pydantic model for enrollment `event` field.

    Note: Only server enrollment statements require an `event` field.

    Attributes:
        course_id (str): Consists in the course in which the student was enrolled or
            unenrolled.
        mode (str): Takes either `audit`, `honor`, `professional` or `verified` value.
            It identifies the student’s enrollment mode.
        user_id (int): Identifies the student who was enrolled or unenrolled.
    """

    course_id: str
    mode: Union[
        Literal["audit"], Literal["honor"], Literal["professional"], Literal["verified"]
    ]
    user_id: Union[int, Literal[""], None]



"""Base xAPI `Object` definitions (2)."""

# Nota bene: we split object definitions into `objects.py` and `unnested_objects.py`
# because of the circular dependency : objects -> context -> objects.

import sys
from datetime import datetime
from typing import List, Optional, Union

from ..config import BaseModelWithConfig
from .agents import BaseXapiAgent
from .attachments import BaseXapiAttachment
from .contexts import BaseXapiContext
from .groups import BaseXapiGroup
from .results import BaseXapiResult
from .unnested_objects import BaseXapiUnnestedObject
from .verbs import BaseXapiVerb

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class BaseXapiSubStatement(BaseModelWithConfig):
    """Pydantic model for `SubStatement` type property.

    Attributes:
        actor (dict): See BaseXapiAgent and BaseXapiGroup.
        verb (dict): See BaseXapiVerb.
        object (dict): See BaseXapiUnnestedObject.
        objectType (dict): Consists of the value `SubStatement`.
    """

    actor: Union[BaseXapiAgent, BaseXapiGroup]
    verb: BaseXapiVerb
    object: BaseXapiUnnestedObject
    objectType: Literal["SubStatement"]
    result: Optional[BaseXapiResult]
    context: Optional[BaseXapiContext]
    timestamp: Optional[datetime]
    attachments: Optional[List[BaseXapiAttachment]]


BaseXapiObject = Union[
    BaseXapiUnnestedObject,
    BaseXapiSubStatement,
    BaseXapiAgent,
    BaseXapiGroup,
]



"""Video xAPI event definitions."""

from ...selector import selector
from ..base.statements import BaseXapiStatement
from ..concepts.activity_types.video import VideoActivity
from ..concepts.verbs.scorm_profile import (
    CompletedVerb,
    InitializedVerb,
    InteractedVerb,
    TerminatedVerb,
)
from ..concepts.verbs.video import PausedVerb, PlayedVerb, SeekedVerb
from .contexts import (
    VideoCompletedContext,
    VideoEnableClosedCaptioningContext,
    VideoInitializedContext,
    VideoPausedContext,
    VideoPlayedContext,
    VideoScreenChangeInteractionContext,
    VideoSeekedContext,
    VideoTerminatedContext,
    VideoVolumeChangeInteractionContext,
)
from .results import (
    VideoCompletedResult,
    VideoEnableClosedCaptioningResult,
    VideoPausedResult,
    VideoPlayedResult,
    VideoScreenChangeInteractionResult,
    VideoSeekedResult,
    VideoTerminatedResult,
    VideoVolumeChangeInteractionResult,
)


class BaseVideoStatement(BaseXapiStatement):
    """Pydantic model for video core statements.

    Attributes:
        object (dict): See VideoActivity.
    """

    object: VideoActivity


class VideoInitialized(BaseVideoStatement):
    """Pydantic model for video initialized statement.

    Example: A video has been fully initialized.

    Attributes:
        verb (dict): See InitializedVerb.
        context (dict): See VideoInitializedContext.
    """

    __selector__ = selector(
        object__definition__type="https://w3id.org/xapi/video/activity-type/video",
        verb__id="http://adlnet.gov/expapi/verbs/initialized",
    )

    verb: InitializedVerb = InitializedVerb()
    context: VideoInitializedContext


class VideoPlayed(BaseVideoStatement):
    """Pydantic model for video played statement.

    Example: John played the video or clicked the play button.

    Attributes:
        verb (dict): See PlayedVerb.
        result (dict): See VideoPlayedResult.
        context (dict): See VideoPlayedContext.
    """

    __selector__ = selector(
        object__definition__type="https://w3id.org/xapi/video/activity-type/video",
        verb__id="https://w3id.org/xapi/video/verbs/played",
    )

    verb: PlayedVerb = PlayedVerb()
    result: VideoPlayedResult
    context: VideoPlayedContext


class VideoPaused(BaseVideoStatement):
    """Pydantic model for video paused statement.

    Example: John paused the video or clicked the pause button.

    Attributes:
        verb (dict): See PausedVerb.
        result (dict): See VideoPausedResult.
        context (dict): See VideoPausedContext.
    """

    __selector__ = selector(
        object__definition__type="https://w3id.org/xapi/video/activity-type/video",
        verb__id="https://w3id.org/xapi/video/verbs/paused",
    )

    verb: PausedVerb = PausedVerb()
    result: VideoPausedResult
    context: VideoPausedContext


class VideoSeeked(BaseVideoStatement):
    """Pydantic model for video seeked statement.

    Example: John moved the progress bar forward or backward to a specific time in the
        video.

    Attributes:
        verb (dict): See SeekedVerb.
        result (dict): See VideoSeekedResult.
        context (dict): See VideoSeekedContext.
    """

    __selector__ = selector(
        object__definition__type="https://w3id.org/xapi/video/activity-type/video",
        verb__id="https://w3id.org/xapi/video/verbs/seeked",
    )

    verb: SeekedVerb = SeekedVerb()
    result: VideoSeekedResult
    context: VideoSeekedContext


class VideoCompleted(BaseVideoStatement):
    """Pydantic model for video completed statement.

    Example: John completed a video by watching major parts of the video at least once.

    Attributes:
        verb (dict): See CompletedVerb.
        result (dict): See VideoCompletedResult.
        context (dict): See VideoCompletedContext.
    """

    __selector__ = selector(
        object__definition__type="https://w3id.org/xapi/video/activity-type/video",
        verb__id="http://adlnet.gov/expapi/verbs/completed",
    )

    verb: CompletedVerb = CompletedVerb()
    result: VideoCompletedResult
    context: VideoCompletedContext


class VideoTerminated(BaseVideoStatement):
    """Pydantic model for video terminated statement.

    Example: John ended a video (quit the player).

    Attributes:
        verb (dict): See TerminatedVerb.
        result (dict): See VideoTerminatedResult.
        context (dict): See VideoTerminatedContext.
    """

    __selector__ = selector(
        object__definition__type="https://w3id.org/xapi/video/activity-type/video",
        verb__id="http://adlnet.gov/expapi/verbs/terminated",
    )

    verb: TerminatedVerb = TerminatedVerb()
    result: VideoTerminatedResult
    context: VideoTerminatedContext


class VideoEnableClosedCaptioning(BaseVideoStatement):
    """Pydantic model for video enable closed captioning statement.

    Example: John interacted with the player to enable closed captioning.

    Attributes:
        verb (dict): See InteractedVerb.
        result (dict): See VideoEnableClosedCaptioningResult.
        context (dict): See VideoEnableClosedCaptioningContext.
    """

    __selector__ = selector(
        object__definition__type="https://w3id.org/xapi/video/activity-type/video",
        verb__id="http://adlnet.gov/expapi/verbs/interacted",
    )

    verb: InteractedVerb = InteractedVerb()
    result: VideoEnableClosedCaptioningResult
    context: VideoEnableClosedCaptioningContext


class VideoVolumeChangeInteraction(BaseVideoStatement):
    """Pydantic model for video volume change interaction statement.

    Example: John interacted with the player to change the volume.

    Attributes :
        verb (dict): See InteractedVerb.
        result (dict): See VideoVolumeChangeInteractionResult.
        context (dict): See VideoVolumeChangeInteractionContext.
    """

    __selector__ = selector(
        object__definition__type="https://w3id.org/xapi/video/activity-type/video",
        verb__id="http://adlnet.gov/expapi/verbs/interacted",
    )

    verb: InteractedVerb = InteractedVerb()
    result: VideoVolumeChangeInteractionResult
    context: VideoVolumeChangeInteractionContext


class VideoScreenChangeInteraction(BaseVideoStatement):
    """Pydantic model for video screen change interaction statement.

    Example: John interacted with the player to activate or deactivate full screen.

    Attributes:
        verb (dict): See InteractedVerb.
        result (dict): See VideoScreenChangeInteractionResult.
        context (dict): See VideoScreenChangeInteractionContext.
    """

    __selector__ = selector(
        object__definition__type="https://w3id.org/xapi/video/activity-type/video",
        verb__id="http://adlnet.gov/expapi/verbs/interacted",
    )

    verb: InteractedVerb = InteractedVerb()
    result: VideoScreenChangeInteractionResult
    context: VideoScreenChangeInteractionContext



"""`Video` verbs definitions."""

import sys
from typing import Dict, Optional

from ...base.verbs import BaseXapiVerb
from ...constants import LANG_EN_US_DISPLAY

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class PlayedVerb(BaseXapiVerb):
    """Pydantic model for played `verb`.

    Attributes:
        id (str): Consists of the value `https://w3id.org/xapi/video/verbs/played`.
        display (dict): Consists of the dictionary `{"en-US": "played"}`.
    """

    id: Literal[
        "https://w3id.org/xapi/video/verbs/played"
    ] = "https://w3id.org/xapi/video/verbs/played"
    display: Optional[Dict[Literal[LANG_EN_US_DISPLAY], Literal["played"]]]


class PausedVerb(BaseXapiVerb):
    """Pydantic model for paused `verb` field.

    Attributes:
        id (str): Consists of the value `https://w3id.org/xapi/video/verbs/paused`.
        display (dict): Consists of the dictionary `{"en-US": "paused"}`.
    """

    id: Literal[
        "https://w3id.org/xapi/video/verbs/paused"
    ] = "https://w3id.org/xapi/video/verbs/paused"
    display: Optional[Dict[Literal[LANG_EN_US_DISPLAY], Literal["paused"]]]


class SeekedVerb(BaseXapiVerb):
    """Pydantic model for seeked `verb` field.

    Attributes:
        id (str): Consists of the value `https://w3id.org/xapi/video/verbs/seeked`.
        display (dict): Consists of the dictionary `{"en-US": "seeked"}`.
    """

    id: Literal[
        "https://w3id.org/xapi/video/verbs/seeked"
    ] = "https://w3id.org/xapi/video/verbs/seeked"
    display: Optional[Dict[Literal[LANG_EN_US_DISPLAY], Literal["seeked"]]]



"""Enrollment event model definitions."""

import sys
from typing import Union

from pydantic import Json

from ralph.models.selector import selector

from ..browser import BaseBrowserModel
from ..server import BaseServerModel
from .fields.contexts import (
    EdxCourseEnrollmentUpgradeClickedContextField,
    EdxCourseEnrollmentUpgradeSucceededContextField,
)
from .fields.events import EnrollmentEventField

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class EdxCourseEnrollmentActivated(BaseServerModel):
    """Pydantic model for `edx.course.enrollment.activated` statement.

    The server emits this statement when a student enrolls in a course.

    Attributes:
        event (EnrollmentEventField): See EnrollmentEventField.
        event_type (str): Consists of the value `edx.course.enrollment.activated`.
        name (str): Consists of the value `edx.course.enrollment.activated`.
    """

    __selector__ = selector(
        event_source="server", event_type="edx.course.enrollment.activated"
    )

    event: Union[
        Json[EnrollmentEventField],
        EnrollmentEventField,
    ]
    event_type: Literal["edx.course.enrollment.activated"]
    name: Literal["edx.course.enrollment.activated"]


class EdxCourseEnrollmentDeactivated(BaseServerModel):
    """Pydantic model for `edx.course.enrollment.deactivated` statement.

    The server emits this statement when a student unenrolls from a course.

    Attributes:
        event (EnrollmentEventField): See EnrollmentEventField.
        event_type (str): Consists of the value `edx.course.enrollment.deactivated`.
        name (str): Consists of the value `edx.course.enrollment.deactivated`.
    """

    __selector__ = selector(
        event_source="server", event_type="edx.course.enrollment.deactivated"
    )

    event: Union[
        Json[EnrollmentEventField],
        EnrollmentEventField,
    ]
    event_type: Literal["edx.course.enrollment.deactivated"]
    name: Literal["edx.course.enrollment.deactivated"]


class EdxCourseEnrollmentModeChanged(BaseServerModel):
    """Pydantic model for `edx.course.enrollment.mode_changed` statement.

    The server emits this statement when the process of changing a student’s
    student_courseenrollment.mode to a different mode is complete.

    Attributes:
        event (EnrollmentEventField): See EnrollmentEventField.
        event_type (str): Consists of the value `edx.course.enrollment.mode_changed`.
        name (str): Consists of the value `edx.course.enrollment.mode_changed`.
    """

    __selector__ = selector(
        event_source="server", event_type="edx.course.enrollment.mode_changed"
    )

    event: Union[
        Json[EnrollmentEventField],
        EnrollmentEventField,
    ]
    event_type: Literal["edx.course.enrollment.mode_changed"]
    name: Literal["edx.course.enrollment.mode_changed"]


class UIEdxCourseEnrollmentUpgradeClicked(BaseBrowserModel):
    """Pydantic model for `edx.course.enrollment.upgrade_clicked` statement.

    The browser emits this statement when a student clicks <kbd>ChallengeYourself</kbd>
    option, and the process of upgrading the student_courseenrollment.mode for the
    student to `verified` begins.

    Attributes:
        context (EdxCourseEnrollmentUpgradeClickedContextField):
            See EdxCourseEnrollmentUpgradeClickedContextField.
        event_type (str): Consists of the value `edx.course.enrollment.upgrade_clicked`.
        name (str): Consists of the value `edx.course.enrollment.upgrade_clicked`.
    """

    __selector__ = selector(
        event_source="browser", event_type="edx.course.enrollment.upgrade_clicked"
    )

    context: EdxCourseEnrollmentUpgradeClickedContextField
    event_type: Literal["edx.course.enrollment.upgrade_clicked"]
    name: Literal["edx.course.enrollment.upgrade_clicked"]


class EdxCourseEnrollmentUpgradeSucceeded(BaseServerModel):
    """Pydantic model for `edx.course.enrollment.upgrade.succeeded` statement.

    The server emits this statement when the process of upgrading a student’s
    student_courseenrollment.mode from `audit` or `honor` to `verified` is complete.

    Attributes:
        context (EdxCourseEnrollmentUpgradeSucceededContextField):
            See EdxCourseEnrollmentUpgradeSucceededContextField.
        event_type (str): Consists of the value
            `edx.course.enrollment.upgrade.succeeded`.
        name (str): Consists of the value `edx.course.enrollment.upgrade.succeeded`.
    """

    __selector__ = selector(
        event_source="server", event_type="edx.course.enrollment.upgrade.succeeded"
    )

    context: EdxCourseEnrollmentUpgradeSucceededContextField
    event_type: Literal["edx.course.enrollment.upgrade.succeeded"]
    name: Literal["edx.course.enrollment.upgrade.succeeded"]



"""FileSystem data backend for Ralph."""

import logging
import os
from datetime import datetime, timezone
from io import BufferedReader, IOBase
from pathlib import Path
from typing import Iterable, Iterator, Optional, Tuple, TypeVar, Union
from uuid import uuid4

from pydantic import PositiveInt

from ralph.backends.data.base import (
    BaseDataBackend,
    BaseDataBackendSettings,
    BaseOperationType,
    DataBackendStatus,
    Listable,
    Writable,
)
from ralph.backends.data.mixins import HistoryMixin
from ralph.conf import BaseSettingsConfig
from ralph.exceptions import BackendException, BackendParameterException
from ralph.utils import now, parse_iterable_to_dict

logger = logging.getLogger(__name__)


class FSDataBackendSettings(BaseDataBackendSettings):
    """FileSystem data backend default configuration.

    Attributes:
        DEFAULT_DIRECTORY_PATH (str or Path): The default target directory path where to
            perform list, read and write operations.
        DEFAULT_QUERY_STRING (str): The default query string to match files for the read
            operation.
        LOCALE_ENCODING (str): The encoding used for writing dictionaries to files.
        READ_CHUNK_SIZE (int): The default chunk size for reading files.
        WRITE_CHUNK_SIZE (int): The default chunk size for writing files.
    """

    class Config(BaseSettingsConfig):
        """Pydantic Configuration."""

        env_prefix = "RALPH_BACKENDS__DATA__FS__"

    DEFAULT_DIRECTORY_PATH: Path = Path(".")
    DEFAULT_QUERY_STRING: str = "*"
    READ_CHUNK_SIZE: int = 4096
    WRITE_CHUNK_SIZE: int = 4096


Settings = TypeVar("Settings", bound=FSDataBackendSettings)


class FSDataBackend(
    BaseDataBackend[Settings, str],
    Writable,
    Listable,
    HistoryMixin,
):
    """FileSystem data backend."""

    name = "fs"
    default_operation_type = BaseOperationType.CREATE
    unsupported_operation_types = {BaseOperationType.DELETE}

    def __init__(self, settings: Optional[Settings] = None):
        """Create the default target directory if it does not exist.

        Args:
            settings (FSDataBackendSettings or None): The data backend settings.
                If `settings` is `None`, a default settings instance is used instead.
        """
        super().__init__(settings)
        self.settings = settings if settings else self.settings_class()
        self.default_directory = self.settings.DEFAULT_DIRECTORY_PATH
        self.default_query_string = self.settings.DEFAULT_QUERY_STRING

        if not self.default_directory.is_dir():
            msg = "Default directory doesn't exist, creating: %s"
            logger.info(msg, self.default_directory)
            self.default_directory.mkdir(parents=True)

        logger.debug("Default directory: %s", self.default_directory)

    def status(self) -> DataBackendStatus:
        """Check whether the default directory has appropriate permissions."""
        for mode in [os.R_OK, os.W_OK, os.X_OK]:
            if not os.access(self.default_directory, mode):
                logger.error(
                    "Invalid permissions for the default directory at %s. "
                    "The directory should have read, write and execute permissions.",
                    str(self.default_directory.absolute()),
                )
                return DataBackendStatus.ERROR

        return DataBackendStatus.OK

    def list(
        self, target: Optional[str] = None, details: bool = False, new: bool = False
    ) -> Union[Iterator[str], Iterator[dict]]:
        """List files and directories in the target directory.

        Args:
            target (str or None): The directory path where to list the files and
                directories.
                If target is `None`, the `default_directory` is used instead.
                If target is a relative path, it is considered to be relative to the
                    `default_directory_path`.
            details (bool): Get detailed file information instead of just file paths.
            new (bool): Given the history, list only not already read files.

        Yield:
            str: The next file path. (If details is False).
            dict: The next file details. (If details is True).

        Raise:
            BackendParameterException: If the `target` argument is not a directory path.
        """
        target: Path = Path(target) if target else self.default_directory
        if not target.is_absolute() and target != self.default_directory:
            target = self.default_directory / target
        try:
            paths = set(target.absolute().iterdir())
        except OSError as error:
            msg = "Invalid target argument"
            logger.error("%s. %s", msg, error)
            raise BackendParameterException(msg, error.strerror) from error

        logger.debug("Found %d files", len(paths))

        if new:
            paths -= set(map(Path, self.get_command_history(self.name, "read")))
            logger.debug("New files: %d", len(paths))

        if not details:
            for path in paths:
                yield str(path)

            return

        for path in paths:
            stats = path.stat()
            modified_at = datetime.fromtimestamp(int(stats.st_mtime), tz=timezone.utc)
            yield {
                "path": str(path),
                "size": stats.st_size,
                "modified_at": modified_at.isoformat(),
            }

    def read(  # noqa: PLR0913
        self,
        query: Optional[str] = None,
        target: Optional[str] = None,
        chunk_size: Optional[int] = None,
        raw_output: bool = False,
        ignore_errors: bool = False,
        max_statements: Optional[PositiveInt] = None,
    ) -> Union[Iterator[bytes], Iterator[dict]]:
        """Read files matching the query in the target folder and yield them.

        Args:
            query (str): The relative pattern for the files to read.
            target (str or None): The target directory path containing the files.
                If target is `None`, the `default_directory_path` is used instead.
                If target is a relative path, it is considered to be relative to the
                    `default_directory_path`.
            chunk_size (int or None): The chunk size when reading files.
                If `chunk_size` is `None` it defaults to `READ_CHUNK_SIZE`.
                If `raw_output` is set to `False`, files are read line by line.
            raw_output (bool): Controls whether to yield bytes or dictionaries.
            ignore_errors (bool): If `True`, encoding errors during the read operation
                will be ignored and logged.
                If `False` (default), a `BackendException` is raised on any error.
            max_statements (int): The maximum number of statements to yield.
                If `None` (default) or `0`, there is no maximum.

        Yield:
            bytes: The next chunk of the read files if `raw_output` is True.
            dict: The next JSON parsed line of the read files if `raw_output` is False.

        Raise:
            BackendException: If a failure during the read operation occurs or
                during JSON encoding lines and `ignore_errors` is set to `False`.
        """
        yield from super().read(
            query, target, chunk_size, raw_output, ignore_errors, max_statements
        )

    def _read_bytes(
        self,
        query: str,
        target: Optional[str],
        chunk_size: int,
        ignore_errors: bool,  # noqa: ARG002
    ) -> Iterator[bytes]:
        """Method called by `self.read` yielding bytes. See `self.read`."""
        for file, path in self._iter_files_matching_query(target, query):
            while chunk := file.read(chunk_size):
                yield chunk
            # The file has been read, add a new entry to the history.
            self._append_to_history("read", path)

    def _read_dicts(
        self,
        query: str,
        target: Optional[str],
        chunk_size: int,  # noqa: ARG002
        ignore_errors: bool,
    ) -> Iterator[dict]:
        """Method called by `self.read` yielding dictionaries. See `self.read`."""
        for file, path in self._iter_files_matching_query(target, query):
            yield from parse_iterable_to_dict(file, ignore_errors)
            # The file has been read, add a new entry to the history.
            self._append_to_history("read", path)

    def _append_to_history(self, action: str, path: Path):
        """Append a new entry to the history."""
        self.append_to_history(
            {
                "backend": self.name,
                "action": action,
                # WARNING: previously only the file name was used as the ID
                # By changing this to the absolute file path, previously fetched
                # files will not be marked as read anymore.
                "id": str(path.absolute()),
                "filename": path.name,
                "size": path.stat().st_size,
                "timestamp": now(),
            }
        )

    def _iter_files_matching_query(
        self, target: Optional[str], query: str
    ) -> Iterator[Tuple[BufferedReader, Path]]:
        """Return file/path tuples for files matching the query in the target folder."""
        if not query:
            query = self.default_query_string

        path = Path(target) if target else self.default_directory
        if not path.is_absolute() and path != self.default_directory:
            path = self.default_directory / path

        paths = list(filter(lambda x: x.is_file(), path.glob(query)))
        if not paths:
            msg = "No file found for query: %s"
            logger.info(msg, path / Path(str(query)))
            return

        logger.debug("Reading matching files: %s", paths)
        for path in paths:
            with path.open("rb") as file:
                yield file, path

    def write(  # noqa: PLR0913
        self,
        data: Union[IOBase, Iterable[bytes], Iterable[dict]],
        target: Optional[str] = None,
        chunk_size: Optional[int] = None,
        ignore_errors: bool = False,
        operation_type: Optional[BaseOperationType] = None,
    ) -> int:
        """Write data records to the target file and return their count.

        Args:
            data (Iterable or IOBase): The data to write.
            target (str or None): The target file path.
                If target is a relative path, it is considered to be relative to the
                    `default_directory_path`.
                If target is `None`, a random (uuid4) file is created in the
                    `default_directory_path` and used as the target instead.
            chunk_size (int or None): Ignored.
            ignore_errors (bool): If `True`, errors during decoding and encoding of
                records are ignored and logged.
                If `False` (default), a `BackendException` is raised on any error.
            operation_type (BaseOperationType or None): The mode of the write operation.
                If operation_type is `CREATE` or `INDEX`, the target file is expected to
                    be absent. If the target file exists a `FileExistsError` is raised.
                If operation_type is `UPDATE`, the target file is overwritten.
                If operation_type is `APPEND`, the data is appended to the
                    end of the target file.

        Return:
            int: The number of written files.

        Raise:
            BackendException: If any failure occurs during the write operation or
                if an inescapable failure occurs and `ignore_errors` is set to `True`.
                E.g.: the `operation_type` is `CREATE` or `INDEX` and the target file
                already exists.
            BackendParameterException: If the `operation_type` is `DELETE` as it is not
                supported.
        """
        return super().write(data, target, chunk_size, ignore_errors, operation_type)

    def _write_dicts(  # noqa: PLR0913
        self,
        data: Iterable[dict],
        target: Optional[str],
        chunk_size: int,
        ignore_errors: bool,
        operation_type: BaseOperationType,
    ) -> int:
        """Method called by `self.write` writing dictionaries. See `self.write`."""
        return super()._write_dicts(
            data, target, chunk_size, ignore_errors, operation_type
        )

    def _write_bytes(  # noqa: PLR0913
        self,
        data: Iterable[bytes],
        target: Optional[str],
        chunk_size: int,  # noqa: ARG002
        ignore_errors: bool,  # noqa: ARG002
        operation_type: BaseOperationType,
    ) -> int:
        """Method called by `self.write` writing bytes. See `self.write`."""
        if not target:
            target = f"{now()}-{uuid4()}"
            msg = "Target file not specified; using random file name: %s"
            logger.info(msg, target)

        path = Path(target)
        path = path if path.is_absolute() else self.default_directory / path

        if operation_type in [BaseOperationType.CREATE, BaseOperationType.INDEX]:
            if path.is_file():
                msg = (
                    "%s already exists and overwrite is not allowed with operation_type"
                    " create or index."
                )
                logger.error(msg, path)
                raise BackendException(msg % path)

            logger.debug("Creating file: %s", path)

        mode = "wb"
        if operation_type == BaseOperationType.APPEND:
            mode = "ab"
            logger.debug("Appending to file: %s", path)

        try:
            with path.open(mode) as file:
                for chunk in data:
                    file.write(chunk)
        except OSError as error:
            msg = "Failed to write to %s: %s"
            logger.error(msg, path, error)
            raise BackendException(msg % (path, error)) from error

        # The file has been created, add a new entry to the history.
        self.append_to_history(
            {
                "backend": self.name,
                "action": "write",
                # WARNING: previously only the file name was used as the ID
                # By changing this to the absolute file path, previously written
                # files will not be marked as written anymore.
                "id": str(path.absolute()),
                "filename": path.name,
                "size": path.stat().st_size,
                "timestamp": now(),
            }
        )
        logger.debug("Written %s with success", path.absolute())
        return 1

    def close(self) -> None:
        """FS backend has no open connections to close. No action."""
        logger.info("No open connections to close; skipping")



"""Utilities for Ralph."""

import asyncio
import datetime
import json
import logging
import operator
from functools import reduce
from importlib import import_module
from inspect import getmembers, isclass, iscoroutine
from itertools import islice
from logging import Logger, getLogger
from typing import (
    Any,
    AsyncIterable,
    AsyncIterator,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
)

from ralph.exceptions import BackendException, UnsupportedBackendException

logger = logging.getLogger(__name__)


def import_subclass(dotted_path: str, parent_class: Any) -> Any:
    """Import a dotted module path.

    Return the class that is a subclass of `parent_class` inside this module.
    Raise ImportError if the import failed.
    """
    module = import_module(dotted_path)

    for _, class_ in getmembers(module, isclass):
        if issubclass(class_, parent_class):
            return class_

    raise ImportError(
        f'Module "{dotted_path}" does not define a subclass of "{parent_class}" class'
    )


# Taken from Django utilities
# https://docs.djangoproject.com/en/3.1/_modules/django/utils/module_loading/#import_string
def import_string(dotted_path: str) -> Any:
    """Import a dotted module path.

    Return the attribute/class designated by the last name in the path.
    Raise ImportError if the import failed.
    """
    try:
        module_path, class_name = dotted_path.rsplit(".", 1)
    except ValueError as err:
        raise ImportError(f"{dotted_path} doesn't look like a module path") from err

    module = import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError as err:
        raise ImportError(
            f'Module "{module_path}" does not define a "{class_name}" attribute/class'
        ) from err


def get_backend_class(backends: Dict[str, Type], name: str) -> Any:
    """Return the backend class from available backends by its name."""
    backend_class = backends.get(name)
    if not backend_class:
        msg = "'%s' backend not found in available backends: %s"
        available_backends = ", ".join(backends.keys())
        logger.error(msg, name, available_backends)
        raise UnsupportedBackendException(msg % (name, available_backends))
    return backend_class


def get_backend_instance(backend_class: Type, options: Dict) -> Any:
    """Return the instantiated backend given the backend class and options."""
    prefix = f"{backend_class.name}_"
    # Filter backend-related parameters. Parameter name is supposed to start
    # with the backend name
    options = {
        name.replace(prefix, "").upper(): value
        for name, value in options.items()
        if name.startswith(prefix) and value is not None
    }
    return backend_class(backend_class.settings_class(**options))


def get_root_logger() -> Logger:
    """Get main Ralph logger."""
    ralph_logger = getLogger("ralph")
    ralph_logger.propagate = True

    return ralph_logger


def now() -> str:
    """Return the current UTC time in ISO format."""
    return datetime.datetime.now(tz=datetime.timezone.utc).isoformat()


def get_dict_value_from_path(dict_: Dict, path: Sequence[str]) -> Union[Dict, None]:
    """Get a nested dictionary value.

    Args:
        dict_ (dict): dictionary of values to which the reduction is
            applied
        path (List): array of keys representing the path to the value
    """
    if path is None:
        return None
    try:
        return reduce(operator.getitem, path, dict_)
    except (KeyError, TypeError):
        return None


def set_dict_value_from_path(dict_: Dict, path: List[str], value: Any) -> None:
    """Set a nested dictionary value.

    Args:
        dict_ (dict): dictionary where the given value is set
        path (List): array of keys representing the path to the value
        value: value to be set
    """
    for key in path[:-1]:
        dict_ = dict_.setdefault(key, {})
    dict_[path[-1]] = value


async def gather_with_limited_concurrency(num_tasks: Optional[int], *tasks: Any) -> Any:
    """Gather no more than `num_tasks` tasks at time.

    Args:
        num_tasks: the maximum number of tasks to run concurrently, if None,
            no maximum value is set
        tasks: tasks to run concurrently
    """
    if num_tasks is not None:
        semaphore = asyncio.Semaphore(num_tasks)

        async def sem_task(task: Any) -> Any:
            async with semaphore:
                return await task

        group = asyncio.gather(*(sem_task(task) for task in tasks))
    else:
        group = asyncio.gather(*tasks)

    # Cancel background tasks on first error
    try:
        return await group
    except Exception as exception:
        group.cancel()
        raise exception


def statements_are_equivalent(statement_1: dict, statement_2: dict) -> bool:
    """Check if statements are equivalent.

    To be equivalent, they must be identical on all fields not modified on input by the
    LRS and identical on other fields, if these fields are present in both
    statements. For example, if an "authority" field is present in only one statement,
    they may still be equivalent.
    """
    # Check that unmutable fields have the same values
    fields = {"actor", "verb", "object", "id", "result", "context", "attachments"}

    # Check that some fields enriched by the LRS are equal when in both statements
    # The LRS specification excludes the fields below from equivalency. It was
    # decided to include them anyway as their value is inherent to the statements.
    other_fields = {"timestamp", "version"}  # "authority" and "stored" remain ignored.
    fields.update(other_fields & statement_1.keys() & statement_2.keys())

    if any(statement_1.get(field) != statement_2.get(field) for field in fields):
        return False
    return True


T = TypeVar("T")


def parse_iterable_to_dict(
    raw_documents: Iterable[T],
    ignore_errors: bool,
    parser: Callable[[T], Dict[str, Any]] = json.loads,
) -> Iterator[dict]:
    """Read the `raw_documents` Iterable and yield dictionaries."""
    for i, raw_document in enumerate(raw_documents):
        try:
            yield parser(raw_document)
        except (TypeError, json.JSONDecodeError) as error:
            msg = "Failed to decode JSON: %s, for document: %s, at line %s"
            if ignore_errors:
                logger.warning(msg, error, raw_document, i)
                continue
            logger.error(msg, error, raw_document, i)
            raise BackendException(msg % (error, raw_document, i)) from error


async def async_parse_iterable_to_dict(
    raw_documents: AsyncIterable[T],
    ignore_errors: bool,
    parser: Callable[[T], Dict[str, Any]] = json.loads,
) -> AsyncIterator[dict]:
    """Read the `raw_documents` Iterable and yield dictionaries."""
    i = 0
    async for raw_document in raw_documents:
        try:
            yield parser(raw_document)
        except (TypeError, json.JSONDecodeError) as error:
            msg = "Failed to decode JSON: %s, for document: %s, at line %s"
            if ignore_errors:
                logger.warning(msg, error, raw_document, i)
                continue
            logger.error(msg, error, raw_document, i)
            raise BackendException(msg % (error, raw_document, i)) from error

        i += 1


def parse_dict_to_bytes(
    documents: Iterable[Dict[str, Any]],
    encoding: str,
    ignore_errors: bool,
) -> Iterator[bytes]:
    """Read the `documents` Iterable with the `encoding` and yield bytes."""
    for i, document in enumerate(documents):
        try:
            yield f"{json.dumps(document)}\n".encode(encoding)
        except (TypeError, ValueError) as error:
            msg = "Failed to encode JSON: %s, for document: %s, at line %s"
            if ignore_errors:
                logger.warning(msg, error, document, i)
                continue
            logger.error(msg, error, document, i)
            raise BackendException(msg % (error, document, i)) from error


async def async_parse_dict_to_bytes(
    documents: AsyncIterable[Dict[str, Any]],
    encoding: str,
    ignore_errors: bool,
) -> AsyncIterator[bytes]:
    """Read the `documents` Iterable with the `encoding` and yield bytes."""
    i = 0
    async for document in documents:
        try:
            yield f"{json.dumps(document)}\n".encode(encoding)
        except (TypeError, ValueError) as error:
            msg = "Failed to encode JSON: %s, for document: %s, at line %s"
            if ignore_errors:
                logger.warning(msg, error, document, i)
                continue
            logger.error(msg, error, document, i)
            raise BackendException(msg % (error, document, i)) from error

        i += 1


def iter_by_batch(iterable: Iterable, n: int) -> Iterable[list]:
    """Batch `iterable` into lists of length `n`. The last batch may be less than `n`.

    Inspired by itertools.batched introduced in python 3.12
    """
    if n < 1:
        raise ValueError("n must be at least one")

    it = iter(iterable)
    while batch := list(islice(it, n)):
        yield batch


def iter_over_async(agenerator) -> Iterable:
    """Iterate synchronously over an asynchronous generator."""
    loop = asyncio.get_event_loop()
    aiterator = agenerator.__aiter__()

    async def get_next():
        """Get the next element from the async iterator."""
        try:
            obj = await aiterator.__anext__()
            return False, obj
        except StopAsyncIteration:
            return True, None

    while True:
        done, obj = loop.run_until_complete(get_next())
        if done:
            break
        yield obj


def execute_async(method):
    """Run asynchronous method in a synchronous context."""

    def wrapper(*args, **kwargs):
        """Wrap method execution."""
        loop = asyncio.get_event_loop()
        loop.run_until_complete(method(*args, **kwargs))

    return wrapper


async def await_if_coroutine(value):
    """Await the value if it is a coroutine, else return synchronously."""
    if iscoroutine(value):
        return await value
    return value



"""Asynchronous MongoDB data backend for Ralph."""

import logging
from io import IOBase
from typing import AsyncIterator, Iterable, Optional, TypeVar, Union

from bson.errors import BSONError
from motor.motor_asyncio import AsyncIOMotorClient, AsyncIOMotorCollection
from pydantic import PositiveInt
from pymongo.errors import BulkWriteError, ConnectionFailure, InvalidName, PyMongoError

from ralph.backends.data.base import BaseOperationType
from ralph.backends.data.mongo import (
    MongoDataBackend,
    MongoDataBackendSettings,
    MongoQuery,
)
from ralph.exceptions import BackendException, BackendParameterException
from ralph.utils import iter_by_batch

from ..data.base import (
    AsyncListable,
    AsyncWritable,
    BaseAsyncDataBackend,
    DataBackendStatus,
)

logger = logging.getLogger(__name__)
Settings = TypeVar("Settings", bound=MongoDataBackendSettings)


class AsyncMongoDataBackend(
    BaseAsyncDataBackend[Settings, MongoQuery],
    AsyncWritable,
    AsyncListable,
):
    """Asynchronous MongoDB data backend."""

    name = "async_mongo"
    unsupported_operation_types = {BaseOperationType.APPEND}

    def __init__(self, settings: Optional[Settings] = None):
        """Instantiate the asynchronous MongoDB client.

        Args:
            settings (MongoDataBackendSettings or None): The data backend settings.
        """
        super().__init__(settings)
        self.client = AsyncIOMotorClient(
            self.settings.CONNECTION_URI, **self.settings.CLIENT_OPTIONS.dict()
        )
        self.database = self.client[self.settings.DEFAULT_DATABASE]
        self.collection = self.database[self.settings.DEFAULT_COLLECTION]

    async def status(self) -> DataBackendStatus:
        """Check the MongoDB connection status.

        Return:
            DataBackendStatus: The status of the data backend.
        """
        # Check MongoDB connection.
        try:
            await self.client.admin.command("ping")
        except (ConnectionFailure, PyMongoError) as error:
            logger.error("Failed to connect to MongoDB: %s", error)
            return DataBackendStatus.AWAY

        # Check MongoDB server status.
        try:
            server_status = await self.client.admin.command("serverStatus")
            if server_status.get("ok") != 1.0:  # noqa: PLR2004
                logger.error("MongoDB `serverStatus` command did not return 1.0")
                return DataBackendStatus.ERROR
        except PyMongoError as error:
            logger.error("Failed to get MongoDB server status: %s", error)
            return DataBackendStatus.ERROR

        return DataBackendStatus.OK

    async def list(
        self, target: Optional[str] = None, details: bool = False, new: bool = False
    ) -> Union[AsyncIterator[str], AsyncIterator[dict]]:
        """List collections in the target database.

        Args:
            target (str or None): The MongoDB database name to list collections from.
                If target is `None`, the `DEFAULT_DATABASE` is used instead.
            details (bool): Get detailed collection information instead of just IDs.
            new (bool): Ignored.

        Yield:
            str: The next collection. (If `details` is False).
            dict: The next collection details. (If `details` is True).

        Raise:
            BackendException: If a failure during the list operation occurs.
            BackendParameterException: If the `target` is not a valid database name.
        """
        if new:
            logger.warning("The `new` argument is ignored")

        try:
            database = self.client[target] if target else self.database
        except InvalidName as error:
            msg = "The target=`%s` is not a valid database name: %s"
            logger.error(msg, target, error)
            raise BackendParameterException(msg % (target, error)) from error

        try:
            collections = await database.list_collections()
            async for collection_info in collections:
                if details:
                    yield collection_info
                else:
                    yield collection_info.get("name")
        except PyMongoError as error:
            msg = "Failed to list MongoDB collections: %s"
            logger.error(msg, error)
            raise BackendException(msg % error) from error

    async def read(  # noqa: PLR0913
        self,
        query: Optional[MongoQuery] = None,
        target: Optional[str] = None,
        chunk_size: Optional[int] = None,
        raw_output: bool = False,
        ignore_errors: bool = False,
        prefetch: Optional[PositiveInt] = None,
        max_statements: Optional[PositiveInt] = None,
    ) -> Union[AsyncIterator[bytes], AsyncIterator[dict]]:
        """Read documents matching the `query` from `target` collection and yield them.

        Args:
            query (MongoQuery): The MongoDB query to use when fetching documents.
            target (str or None): The MongoDB collection name to query.
                If target is `None`, the `DEFAULT_COLLECTION` is used instead.
            chunk_size (int or None): The chunk size when reading documents by batches.
                If `chunk_size` is `None` it defaults to `READ_CHUNK_SIZE`.
            raw_output (bool): Whether to yield dictionaries or bytes.
            ignore_errors (bool): If `True`, encoding errors during the read operation
                will be ignored and logged.
                If `False` (default), a `BackendException` is raised on any error.
            prefetch (int): The number of records to prefetch (queue) while yielding.
                If `prefetch` is `None` it defaults to `1`, i.e. no records are
                prefetched.
            max_statements (int): The maximum number of statements to yield.
                If `None` (default) or `0`, there is no maximum.

        Yield:
            bytes: The next raw document if `raw_output` is True.
            dict: The next JSON parsed document if `raw_output` is False.

        Raise:
            BackendException: If a failure occurs during MongoDB connection or
                during encoding documents and `ignore_errors` is set to `False`.
            BackendParameterException: If the `target` is not a valid collection name.
        """
        statements = super().read(
            query,
            target,
            chunk_size,
            raw_output,
            ignore_errors,
            prefetch,
            max_statements,
        )
        async for statement in statements:
            yield statement

    async def _read_dicts(
        self,
        query: MongoQuery,
        target: Optional[str],
        chunk_size: int,
        ignore_errors: bool,  # noqa: ARG002
    ) -> AsyncIterator[dict]:
        """Method called by `self.read` yielding dictionaries. See `self.read`."""
        kwargs = query.dict(exclude_unset=True)
        collection = self._get_target_collection(target)
        try:
            async for document in collection.find(batch_size=chunk_size, **kwargs):
                document.update({"_id": str(document.get("_id"))})
                yield document
        except (PyMongoError, IndexError, TypeError, ValueError) as error:
            msg = "Failed to execute MongoDB query: %s"
            logger.error(msg, error)
            raise BackendException(msg % error) from error

    async def write(  # noqa: PLR0913
        self,
        data: Union[IOBase, Iterable[bytes], Iterable[dict]],
        target: Optional[str] = None,
        chunk_size: Optional[int] = None,
        ignore_errors: bool = False,
        operation_type: Optional[BaseOperationType] = None,
        concurrency: Optional[PositiveInt] = None,
    ) -> int:
        """Write data documents to the target collection and return their count.

        Args:
            data (Iterable or IOBase): The data containing documents to write.
            target (str or None): The target MongoDB collection name.
            chunk_size (int or None): The number of documents to write in one batch.
                If `chunk_size` is `None` it defaults to `WRITE_CHUNK_SIZE`.
            ignore_errors (bool): If `True`, errors during decoding, encoding and
                sending batches of documents are ignored and logged.
                If `False` (default), a `BackendException` is raised on any error.
            operation_type (BaseOperationType or None): The mode of the write operation.
                If `operation_type` is `None`, the `default_operation_type` is used
                    instead. See `BaseOperationType`.
            concurrency (int): The number of chunks to write concurrently.
                If `None` it defaults to `1`.

        Return:
            int: The number of documents written.

        Raise:
            BackendException: If any failure occurs during the write operation or
                if an inescapable failure occurs and `ignore_errors` is set to `True`.
            BackendParameterException: If the `operation_type` is `APPEND` as it is not
                supported.
        """
        return await super().write(
            data, target, chunk_size, ignore_errors, operation_type, concurrency
        )

    async def _write_dicts(  # noqa: PLR0913
        self,
        data: Iterable[dict],
        target: Optional[str],
        chunk_size: int,
        ignore_errors: bool,
        operation_type: BaseOperationType,
    ) -> int:
        """Method called by `self.write` writing dictionaries. See `self.write`."""
        count = 0
        collection = self._get_target_collection(target)
        msg = "Start writing to the %s collection of the %s database (chunk size: %d)"
        logger.debug(msg, collection, self.database, chunk_size)
        if operation_type == BaseOperationType.UPDATE:
            data = MongoDataBackend.to_replace_one(data)
            for batch in iter_by_batch(data, chunk_size):
                count += await self._bulk_update(batch, ignore_errors, collection)
            logger.info("Updated %d documents with success", count)
        elif operation_type == BaseOperationType.DELETE:
            for batch in iter_by_batch(MongoDataBackend.to_ids(data), chunk_size):
                count += await self._bulk_delete(batch, ignore_errors, collection)
            logger.info("Deleted %d documents with success", count)
        else:
            data = MongoDataBackend.to_documents(data, ignore_errors, operation_type)
            for batch in iter_by_batch(data, chunk_size):
                count += await self._bulk_import(batch, ignore_errors, collection)
            logger.info("Inserted %d documents with success", count)

        return count

    async def close(self) -> None:
        """Close the AsyncIOMotorClient client.

        Raise:
            BackendException: If a failure occurs during the close operation.
        """
        try:
            self.client.close()
        except PyMongoError as error:
            msg = "Failed to close AsyncIOMotorClient: %s"
            logger.error(msg, error)
            raise BackendException(msg % error) from error

    def _get_target_collection(
        self, target: Union[str, None]
    ) -> AsyncIOMotorCollection:
        """Return the target collection."""
        try:
            return self.database[target] if target else self.collection
        except InvalidName as error:
            msg = "The target=`%s` is not a valid collection name: %s"
            logger.error(msg, target, error)
            raise BackendParameterException(msg % (target, error)) from error

    async def _bulk_import(
        self, batch: list, ignore_errors: bool, collection: AsyncIOMotorCollection
    ):
        """Insert a batch of documents into the selected database collection."""
        try:
            new_documents = await collection.insert_many(batch)
        except (BulkWriteError, PyMongoError, BSONError, ValueError) as error:
            msg = "Failed to insert document chunk: %s"
            if ignore_errors:
                logger.warning(msg, error)
                return getattr(error, "details", {}).get("nInserted", 0)
            raise BackendException(msg % error) from error

        inserted_count = len(new_documents.inserted_ids)
        logger.debug("Inserted %d documents chunk with success", inserted_count)
        return inserted_count

    async def _bulk_delete(
        self, batch: list, ignore_errors: bool, collection: AsyncIOMotorCollection
    ):
        """Delete a batch of documents from the selected database collection."""
        try:
            deleted_documents = await collection.delete_many(
                {"_source.id": {"$in": batch}}
            )
        except (BulkWriteError, PyMongoError, BSONError, ValueError) as error:
            msg = "Failed to delete document chunk: %s"
            if ignore_errors:
                logger.warning(msg, error)
                return getattr(error, "details", {}).get("nRemoved", 0)
            raise BackendException(msg % error) from error

        deleted_count = deleted_documents.deleted_count
        logger.debug("Deleted %d documents chunk with success", deleted_count)
        return deleted_count

    async def _bulk_update(
        self, batch: list, ignore_errors: bool, collection: AsyncIOMotorCollection
    ):
        """Update a batch of documents into the selected database collection."""
        try:
            updated_documents = await collection.bulk_write(batch)
        except (BulkWriteError, PyMongoError, BSONError, ValueError) as error:
            msg = "Failed to update document chunk: %s"
            if ignore_errors:
                logger.warning(msg, error)
                return getattr(error, "details", {}).get("nModified", 0)
            logger.error(msg, error)
            raise BackendException(msg % error) from error

        modified_count = updated_documents.modified_count
        logger.debug("Updated %d documents chunk with success", modified_count)
        return modified_count



#!/usr/bin/env python

"""Ralph main entrypoint.

This script is a POC for now, we aim to make ralph a nice CLI to use.
"""

import sentry_sdk

from ralph.conf import settings

from . import __version__, cli

if settings.SENTRY_DSN is not None:
    sentry_sdk.init(
        dsn=settings.SENTRY_DSN,
        traces_sample_rate=settings.SENTRY_CLI_TRACES_SAMPLE_RATE,
        release=__version__,
        environment=settings.EXECUTION_ENVIRONMENT,
        max_breadcrumbs=50,
    )

if __name__ == "__main__":
    cli.cli()



"""`Scorm Profile` activity types definitions."""

import sys

from ...base.unnested_objects import BaseXapiActivity, BaseXapiActivityDefinition

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


# CMI Interaction


class CMIInteractionActivityDefinition(BaseXapiActivityDefinition):
    """Pydantic model for CMI Interaction `Activity` type `definition` property.

    Attributes:
        type (str): Consists of the value
            `http://adlnet.gov/expapi/activities/cmi.interaction`.
    """

    type: Literal[
        "http://adlnet.gov/expapi/activities/cmi.interaction"
    ] = "http://adlnet.gov/expapi/activities/cmi.interaction"


class CMIInteractionActivity(BaseXapiActivity):
    """Pydantic model for CMI Interaction `Activity` type.

    Attributes:
        definition (dict): see CMIInteractionActivityDefinition.
    """

    definition: CMIInteractionActivityDefinition = CMIInteractionActivityDefinition()


# Profile


class ProfileActivityDefinition(BaseXapiActivityDefinition):
    """Pydantic model for profile `Activity` type `definition` property.

    Attributes:
        type (str): Consists of the value
            `http://adlnet.gov/expapi/activities/profile`.
    """

    type: Literal[
        "http://adlnet.gov/expapi/activities/profile"
    ] = "http://adlnet.gov/expapi/activities/profile"


class ProfileActivity(BaseXapiActivity):
    """Pydantic model for profile `Activity` type.

    Attributes:
        definition (dict): see ProfileActivityDefinition.
    """

    definition: ProfileActivityDefinition = ProfileActivityDefinition()


# Course


class CourseActivityDefinition(BaseXapiActivityDefinition):
    """Pydantic model for course `Activity` type `definition` property.

    Attributes:
        type (str): Consists of the value
            `http://adlnet.gov/expapi/activities/course`.
    """

    type: Literal[
        "http://adlnet.gov/expapi/activities/course"
    ] = "http://adlnet.gov/expapi/activities/course"


class CourseActivity(BaseXapiActivity):
    """Pydantic model for course `Activity` type.

    Attributes:
        definition (dict): see CourseActivityDefinition.
    """

    definition: CourseActivityDefinition = CourseActivityDefinition()


# Module


class ModuleActivityDefinition(BaseXapiActivityDefinition):
    """Pydantic model for module `Activity` type `definition` property.

    Attributes:
        type (str): Consists of the value
            `http://adlnet.gov/expapi/activities/module`.
    """

    type: Literal[
        "http://adlnet.gov/expapi/activities/module"
    ] = "http://adlnet.gov/expapi/activities/module"


class ModuleActivity(BaseXapiActivity):
    """Pydantic model for module `Activity` type.

    Attributes:
        definition (dict): see ModuleActivityDefinition.
    """

    definition: ModuleActivityDefinition = ModuleActivityDefinition()



"""Textbook interaction event model definitions."""

import sys
from typing import Union

from pydantic import Json

from ralph.models.selector import selector

from ..browser import BaseBrowserModel
from .fields.events import (
    BookEventField,
    TextbookPdfChapterNavigatedEventField,
    TextbookPdfDisplayScaledEventField,
    TextbookPdfOutlineToggledEventField,
    TextbookPdfPageNavigatedEventField,
    TextbookPdfPageScrolledEventField,
    TextbookPdfSearchCaseSensitivityToggledEventField,
    TextbookPdfSearchExecutedEventField,
    TextbookPdfSearchHighlightToggledEventField,
    TextbookPdfSearchNavigatedNextEventField,
    TextbookPdfThumbnailNavigatedEventField,
    TextbookPdfThumbnailsToggledEventField,
    TextbookPdfZoomButtonsChangedEventField,
    TextbookPdfZoomMenuChangedEventField,
)

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class UIBook(BaseBrowserModel):
    """Pydantic model for `book` statement.

    The browser emits this statement when a user navigates within the PDF Viewer or the
    PNG Viewer.

    Attributes:
        event (BookEventField): See BookEventField.
        event_type (str): Consists of the value `book`.
        name (str): Consists of the value `book`.
    """

    __selector__ = selector(event_source="browser", event_type="book")

    event: Union[Json[BookEventField], BookEventField]
    event_type: Literal["book"]
    name: Literal["book"]


class UITextbookPdfThumbnailsToggled(BaseBrowserModel):
    """Pydantic model for `textbook.pdf.thumbnails.toggled` statement.

    The browser emits this statement when a user clicks on the icon to show or hide
    page thumbnails.

    Attributes:
        event (json): See TextbookPdfThumbnailsToggledEventField.
        event_type (str): Consists of the value `textbook.pdf.thumbnails.toggled`.
        name (str): Consists of the value `textbook.pdf.thumbnails.toggled`.
    """

    __selector__ = selector(
        event_source="browser", event_type="textbook.pdf.thumbnails.toggled"
    )

    event: Union[
        Json[TextbookPdfThumbnailsToggledEventField],
        TextbookPdfThumbnailsToggledEventField,
    ]
    event_type: Literal["textbook.pdf.thumbnails.toggled"]
    name: Literal["textbook.pdf.thumbnails.toggled"]


class UITextbookPdfThumbnailNavigated(BaseBrowserModel):
    """Pydantic model for `textbook.pdf.thumbnail.navigated` statement.

    The browser emits this statement when a user clicks on a thumbnail image to
    navigate to a page.

    Attributes:
        event (json): See TextbookPdfThumbnailNavigatedEventField.
        event_type (str): Consists of the value `textbook.pdf.thumbnail.navigated`.
        name (str): Consists of the value `textbook.pdf.thumbnail.navigated`.
    """

    __selector__ = selector(
        event_source="browser", event_type="textbook.pdf.thumbnail.navigated"
    )

    event: Union[
        Json[TextbookPdfThumbnailNavigatedEventField],
        TextbookPdfThumbnailNavigatedEventField,
    ]
    event_type: Literal["textbook.pdf.thumbnail.navigated"]
    name: Literal["textbook.pdf.thumbnail.navigated"]


class UITextbookPdfOutlineToggled(BaseBrowserModel):
    """Pydantic model for `textbook.pdf.outline.toggled` statement.

    The browser emits this statement when a user clicks the outline icon to show or
    hide a list of the book’s chapters.

    Attributes:
        event (json): See TextbookPdfOutlineToggledEventField.
        event_type (str): Consists of the value `textbook.pdf.outline.toggled`.
        name (str): Consists of the value `textbook.pdf.outline.toggled`.
    """

    __selector__ = selector(
        event_source="browser", event_type="textbook.pdf.outline.toggled"
    )

    event: Union[
        Json[TextbookPdfOutlineToggledEventField],
        TextbookPdfOutlineToggledEventField,
    ]
    event_type: Literal["textbook.pdf.outline.toggled"]
    name: Literal["textbook.pdf.outline.toggled"]


class UITextbookPdfChapterNavigated(BaseBrowserModel):
    """Pydantic model for `textbook.pdf.chapter.navigated` statement.

    The browser emits this statement when a user clicks on a link in the outline to
    navigate to a chapter.

    Attributes:
        event (json): See TextbookPdfChapterNavigatedEventField.
        event_type (str): Consists of the value `textbook.pdf.chapter.navigated`.
        name (str): Consists of the value `textbook.pdf.chapter.navigated`.
    """

    __selector__ = selector(
        event_source="browser", event_type="textbook.pdf.chapter.navigated"
    )

    event: Union[
        Json[TextbookPdfChapterNavigatedEventField],
        TextbookPdfChapterNavigatedEventField,
    ]
    event_type: Literal["textbook.pdf.chapter.navigated"]
    name: Literal["textbook.pdf.chapter.navigated"]


class UITextbookPdfPageNavigated(BaseBrowserModel):
    """Pydantic model for `textbook.pdf.page.navigated` statement.

    The browser emits this statement when a user manually enters a page number.

    Attributes:
        event (json): See TextbookPdfPageNavigatedEventField.
        event_type (str): Consists of the value `textbook.pdf.page.navigated`.
        name (str): Consists of the value `textbook.pdf.page.navigated`.
    """

    __selector__ = selector(
        event_source="browser", event_type="textbook.pdf.page.navigated"
    )

    event: Union[
        Json[TextbookPdfPageNavigatedEventField],
        TextbookPdfPageNavigatedEventField,
    ]
    event_type: Literal["textbook.pdf.page.navigated"]
    name: Literal["textbook.pdf.page.navigated"]


class UITextbookPdfZoomButtonsChanged(BaseBrowserModel):
    """Pydantic model for `textbook.pdf.zoom.buttons.changed` statement.

    The browser emits this statement when a user clicks either the <kbd>Zoom In</kbd>
    or <kbd>Zoom Out</kbd> icon.

    Attributes:
        event (json): See TextbookPdfZoomButtonsChangedEventField.
        event_type (str): Consists of the value `textbook.pdf.zoom.buttons.changed`.
        name (str): Consists of the value `textbook.pdf.zoom.buttons.changed`.
    """

    __selector__ = selector(
        event_source="browser", event_type="textbook.pdf.zoom.buttons.changed"
    )

    event: Union[
        Json[TextbookPdfZoomButtonsChangedEventField],
        TextbookPdfZoomButtonsChangedEventField,
    ]
    event_type: Literal["textbook.pdf.zoom.buttons.changed"]
    name: Literal["textbook.pdf.zoom.buttons.changed"]


class UITextbookPdfZoomMenuChanged(BaseBrowserModel):
    """Pydantic model for `textbook.pdf.zoom.menu.changed` statement.

    The browser emits this statement when a user selects a magnification setting.

    Attributes:
        event (json): See TextbookPdfZoomMenuChangedEventField.
        event_type (str): Consists of the value `textbook.pdf.zoom.menu.changed`.
        name (str): Consists of the value `textbook.pdf.zoom.menu.changed`.
    """

    __selector__ = selector(
        event_source="browser", event_type="textbook.pdf.zoom.menu.changed"
    )

    event: Union[
        Json[TextbookPdfZoomMenuChangedEventField],
        TextbookPdfZoomMenuChangedEventField,
    ]
    event_type: Literal["textbook.pdf.zoom.menu.changed"]
    name: Literal["textbook.pdf.zoom.menu.changed"]


class UITextbookPdfDisplayScaled(BaseBrowserModel):
    """Pydantic model for `textbook.pdf.display.scaled` statement.

    The browser emits this statement when the display magnification changes or the
    first page is shown.

    Attributes:
        event (json): See TextbookPdfDisplayScaledEventField.
        event_type (str): Consists of the value `textbook.pdf.display.scaled`.
        name (str): Consists of the value `textbook.pdf.display.scaled`.
    """

    __selector__ = selector(
        event_source="browser", event_type="textbook.pdf.display.scaled"
    )

    event: Union[
        Json[TextbookPdfDisplayScaledEventField],
        TextbookPdfDisplayScaledEventField,
    ]
    event_type: Literal["textbook.pdf.display.scaled"]
    name: Literal["textbook.pdf.display.scaled"]


class UITextbookPdfPageScrolled(BaseBrowserModel):
    """Pydantic model for `textbook.pdf.page.scrolled` statement.

    The browser emits this statement when the user scrolls to the next or previous page
    and the transition takes less than 50 milliseconds.

    Attributes:
        event (json): See TextbookPdfPageScrolledEventField.
        event_type (str): Consists of the value `textbook.pdf.page.scrolled`.
        name (str): Consists of the value `textbook.pdf.page.scrolled`.
    """

    __selector__ = selector(
        event_source="browser", event_type="textbook.pdf.page.scrolled"
    )

    event: Union[
        Json[TextbookPdfPageScrolledEventField],
        TextbookPdfPageScrolledEventField,
    ]
    event_type: Literal["textbook.pdf.page.scrolled"]
    name: Literal["textbook.pdf.page.scrolled"]


class UITextbookPdfSearchExecuted(BaseBrowserModel):
    """Pydantic model for `textbook.pdf.search.executed` statement.

    The browser emits this statement when a user searches for a text value in the file.

    Attributes:
        event (json): See TextbookPdfSearchExecutedEventField.
        event_type (str): Consists of the value `textbook.pdf.search.executed`.
        name (str): Consists of the value `textbook.pdf.search.executed`.
    """

    __selector__ = selector(
        event_source="browser", event_type="textbook.pdf.search.executed"
    )

    event: Union[
        Json[TextbookPdfSearchExecutedEventField],
        TextbookPdfSearchExecutedEventField,
    ]
    event_type: Literal["textbook.pdf.search.executed"]
    name: Literal["textbook.pdf.search.executed"]


class UITextbookPdfSearchNavigatedNext(BaseBrowserModel):
    """Pydantic model for `textbook.pdf.search.navigatednext` statement.

    The browser emits this statement when a user clicks on the <kbd>Find Next</kbd> or
    <kbd>Find Previous</kbd> icons for an entered search string.

    Attributes:
        event (json): See TextbookPdfSearchNavigatedNextEventField.
        event_type (str): Consists of the value `textbook.pdf.search.navigatednext`.
        name (str): Consists of the value `textbook.pdf.search.navigatednext`.
    """

    __selector__ = selector(
        event_source="browser", event_type="textbook.pdf.search.navigatednext"
    )

    event: Union[
        Json[TextbookPdfSearchNavigatedNextEventField],
        TextbookPdfSearchNavigatedNextEventField,
    ]
    event_type: Literal["textbook.pdf.search.navigatednext"]
    name: Literal["textbook.pdf.search.navigatednext"]


class UITextbookPdfSearchHighlightToggled(BaseBrowserModel):
    """Pydantic model for `textbook.pdf.search.highlight.toggled` statement.

    The browser emits this statement when a user selects or clears the
    <kbd>Highlight All</kbd> option.

    Attributes:
        event (json): See TextbookPdfSearchHighlightToggledEventField.
        event_type (str): Consists of the value `textbook.pdf.search.highlight.toggled`.
        name (str): Consists of the value `textbook.pdf.search.highlight.toggled`.
    """

    __selector__ = selector(
        event_source="browser", event_type="textbook.pdf.search.highlight.toggled"
    )

    event: Union[
        Json[TextbookPdfSearchHighlightToggledEventField],
        TextbookPdfSearchHighlightToggledEventField,
    ]
    event_type: Literal["textbook.pdf.search.highlight.toggled"]
    name: Literal["textbook.pdf.search.highlight.toggled"]


class UITextbookPdfSearchCaseSensitivityToggled(BaseBrowserModel):
    """Pydantic model for `textbook.pdf.searchcasesensitivity.toggled` statement.

    The browser emits this statement when a user selects or clears the
    <kbd>Match Case</kbd> option.

    Attributes:
        event (json): See TextbookPdfSearchCaseSensitivityToggledEventField.
        event_type (str): Consists of the value
            `textbook.pdf.searchcasesensitivity.toggled`.
        name (str): Consists of the value `textbook.pdf.searchcasesensitivity.toggled`.
    """

    __selector__ = selector(
        event_source="browser", event_type="textbook.pdf.searchcasesensitivity.toggled"
    )

    event: Union[
        Json[TextbookPdfSearchCaseSensitivityToggledEventField],
        TextbookPdfSearchCaseSensitivityToggledEventField,
    ]
    event_type: Literal["textbook.pdf.searchcasesensitivity.toggled"]
    name: Literal["textbook.pdf.searchcasesensitivity.toggled"]



"""Ralph logging manager."""

from logging.config import dictConfig

from ralph.conf import settings
from ralph.exceptions import ConfigurationException


def configure_logging() -> None:
    """Set up Ralph logging configuration."""
    try:
        dictConfig(settings.LOGGING)
    except Exception as error:  # noqa: BLE001
        raise ConfigurationException("Improperly configured logging") from error



"""Main module for Ralph's LRS API authentication."""
from typing import Optional

from fastapi import Depends, HTTPException, status
from fastapi.security import SecurityScopes

from ralph.api.auth.basic import AuthenticatedUser, get_basic_auth_user
from ralph.api.auth.oidc import get_oidc_user
from ralph.conf import AuthBackend, settings


def get_authenticated_user(
    security_scopes: SecurityScopes = SecurityScopes([]),
    basic_auth_user: Optional[AuthenticatedUser] = Depends(get_basic_auth_user),
    oidc_auth_user: Optional[AuthenticatedUser] = Depends(get_oidc_user),
) -> AuthenticatedUser:
    """Authenticate user with any allowed method, using credentials in the header."""
    if AuthBackend.BASIC not in settings.RUNSERVER_AUTH_BACKENDS:
        basic_auth_user = None
    if AuthBackend.OIDC not in settings.RUNSERVER_AUTH_BACKENDS:
        oidc_auth_user = None

    if basic_auth_user:
        user = basic_auth_user
        auth_header = "Basic"
    elif oidc_auth_user:
        user = oidc_auth_user
        auth_header = "Bearer"
    else:
        auth_header = ",".join(
            [
                {"basic": "Basic", "oidc": "Bearer"}[backend.value]
                for backend in settings.RUNSERVER_AUTH_BACKENDS
            ]
        )
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Invalid authentication credentials",
            headers={"WWW-Authenticate": auth_header},
        )

    # Restrict access by scopes
    if settings.LRS_RESTRICT_BY_SCOPES:
        for requested_scope in security_scopes.scopes:
            if not user.scopes.is_authorized(requested_scope):
                raise HTTPException(
                    status_code=status.HTTP_401_UNAUTHORIZED,
                    detail=f'Access not authorized to scope: "{requested_scope}".',
                    headers={"WWW-Authenticate": auth_header},
                )
    return user



"""Server event model definitions."""

import sys
from pathlib import Path
from typing import Union

from pydantic import Json

from ralph.models.selector import LazyModelField, selector

from .base import AbstractBaseEventField, BaseEdxModel

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class BaseServerModel(BaseEdxModel):
    """Pydantic model for core server statement."""

    event_source: Literal["server"]


class ServerEventField(AbstractBaseEventField):
    """Pydantic model for common server `event` field."""

    GET: dict
    POST: dict


class Server(BaseServerModel):
    """Pydantic model for common server statement.

    This type of event is triggered from the django middleware on each request
    excluding: `/event`, `login`, `heartbeat`, `/segmentio/event` and `/performance`.

    Attributes:
        event_type (str): Consist of the relative URL (without the hostname) of the
            requested page.
            Retrieved with:
                `request.META['PATH_INFO']`
        event (str): Consist of a JSON string holding the content of the GET or POST
            request.
            Retrieved with:
                ```json.dumps(
                    {
                        'GET': dict(request.GET),
                        'POST': dict(request.POST)
                    }
                )[:512]```
            Note:
                Values for ['password', 'newpassword', 'new_password', 'oldpassword',
                'old_password', 'new_password1', 'new_password2'] are replaced by
                `********`.
                The JSON string is truncated at 512 characters resulting in invalid
                JSON.
    """

    __selector__ = selector(
        event_source="server", event_type=LazyModelField("context__path")
    )

    event_type: Path
    event: Union[Json[ServerEventField], ServerEventField]



"""Base xAPI `Verb` definitions."""

from typing import Optional

from ..config import BaseModelWithConfig
from .common import IRI, LanguageMap


class BaseXapiVerb(BaseModelWithConfig):
    """Pydantic model for `verb` property.

    Attributes:
        id (IRI): Consists of an identifier for the verb.
        display (LanguageMap): Consists of a human-readable representation of the verb.
    """

    id: IRI
    display: Optional[LanguageMap]



"""Base xAPI `Inverse Functional Identifier` definitions."""

from pydantic import AnyUrl, StrictStr, constr

from ..config import BaseModelWithConfig
from .common import IRI, MailtoEmail


class BaseXapiAccount(BaseModelWithConfig):
    """Pydantic model for IFI `account` property.

    Attributes:
        homePage (IRI): Consists of the home page of the account's service provider.
        name (str): Consists of the unique id or name of the Actor's account.
    """

    homePage: IRI
    name: StrictStr


class BaseXapiMboxIFI(BaseModelWithConfig):
    """Pydantic model for mailto Inverse Functional Identifier.

    Attributes:
        mbox (MailtoEmail): Consists of the Agent's email address.
    """

    mbox: MailtoEmail


class BaseXapiMboxSha1SumIFI(BaseModelWithConfig):
    """Pydantic model for hash Inverse Functional Identifier.

    Attributes:
        mbox_sha1sum (str): Consists of the SHA1 hash of the Agent's email address.
    """

    mbox_sha1sum: constr(regex=r"^[0-9a-f]{40}$")


class BaseXapiOpenIdIFI(BaseModelWithConfig):
    """Pydantic model for OpenID Inverse Functional Identifier.

    Attributes:
        openid (URI): Consists of an openID that uniquely identifies the Agent.
    """

    openid: AnyUrl


class BaseXapiAccountIFI(BaseModelWithConfig):
    """Pydantic model for account Inverse Functional Identifier.

    Attributes:
        account (dict): See BaseXapiAccount.
    """

    account: BaseXapiAccount



"""Ralph tracking logs parsers."""

import json
import logging
from abc import ABC, abstractmethod
from typing import BinaryIO, Generator, TextIO, Union

logger = logging.getLogger(__name__)


class BaseParser(ABC):
    """Base tracking logs parser."""

    name = "base"

    @abstractmethod
    def parse(self, input_file: Union[TextIO, BinaryIO]) -> Generator:
        """Parse GELF formatted logs (one JSON string event per row).

        Args:
            input_file (file-like): The log file to parse.

        Yields:
            event: raw event as extracted from its container.
        """


class GELFParser(BaseParser):
    """GELF formatted logs parser.

    Documentation of the GELF log format can be found in the Graylog project's
    documentation: https://docs.graylog.org/en/latest/pages/gelf.html
    """

    name = "gelf"

    def parse(self, input_file: Union[TextIO, BinaryIO]) -> Generator:
        """Parse GELF formatted logs (one JSON string event per row).

        Args:
            input_file (file-like): The log file to parse.

        Yields:
            event: Events raw short_message string.
        """
        logger.info("Parsing: %s", input_file)

        for event in input_file:
            try:
                yield json.loads(event)["short_message"]
            except (json.JSONDecodeError, TypeError) as err:
                msg = "Input event '%s' is not a valid JSON string! It will be ignored."
                logger.error(msg, event)
                logger.debug("Raised error was: %s", err)
            except KeyError as err:
                msg = (
                    "Input event '%s' doesn't comply with GELF format! "
                    "It will be ignored."
                )
                logger.error(msg, event)
                logger.debug("Raised error was: %s", err)


class ElasticSearchParser(BaseParser):
    """ElasticSearch JSON document parser."""

    name = "es"

    def parse(self, input_file: Union[TextIO, BinaryIO]) -> Generator:
        """Parse Elasticsearch JSON documents.

        Args:
            input_file (file-like): The file containing Elasticsearch JSON documents.

        Yields:
            document: ElasticSearch documents `_source` field content.
        """
        logger.info("Parsing: %s", input_file)

        for document in input_file:
            try:
                yield json.dumps(json.loads(document)["_source"])
            except (json.JSONDecodeError, TypeError) as err:
                msg = "Document '%s' is not a valid JSON string! It will be ignored."
                logger.error(msg, document)
                logger.debug("Raised error was: %s", err)
            except KeyError as err:
                msg = "Document '%s' has no `_source` field! It will be ignored."
                logger.error(msg, document)
                logger.debug("Raised error was: %s", err)



"""LMS xAPI events object fields definitions."""

import sys
from typing import Optional

from pydantic import Field

from ..concepts.activity_types.acrossx_profile import (
    WebpageActivity,
    WebpageActivityDefinition,
)
from ..concepts.activity_types.activity_streams_vocabulary import (
    FileActivity,
    FileActivityDefinition,
)
from ..concepts.constants.acrossx_profile import ACTIVITY_EXTENSIONS_TYPE
from ..config import BaseExtensionModelWithConfig

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


# Page


class LMSPageObjectDefinitionExtensions(BaseExtensionModelWithConfig):
    """Pydantic model for LMS page `object`.`definition`.`extensions` property.

    Attributes:
        type (str): Characterisation of the page. Can be either `course`,
            `course_list`, `user_space` value.
    """

    type: Optional[Literal["course", "course_list", "user_space"]] = Field(
        alias=ACTIVITY_EXTENSIONS_TYPE
    )


class LMSPageObjectDefinition(WebpageActivityDefinition):
    """Pydantic model for LMS page `object`.`definition` property.

    Attributes:
        extensions (dict): see LMSPageObjectDefinitionExtensions.
    """

    extensions: Optional[LMSPageObjectDefinitionExtensions]


class LMSPageObject(WebpageActivity):
    """Pydantic model for LMS page `object` property.

    Attributes:
        definition (dict): see LMSPageObjectDefinition.
    """

    definition: LMSPageObjectDefinition


# File


class LMSFileObjectDefinitionExtensions(BaseExtensionModelWithConfig):
    """Pydantic model for LMS file `object`.`definition`.`extensions` property.

    Attributes:
        type (str): Characterisation of the MIME type of the file.
    """

    type: str = Field(alias=ACTIVITY_EXTENSIONS_TYPE)


class LMSFileObjectDefinition(FileActivityDefinition):
    """Pydantic model for LMS file `object`.`definition` property.

    Attributes:
        extensions (dict): see LMSFileObjectDefinitionExtensions.
    """

    extensions: Optional[LMSFileObjectDefinitionExtensions]


class LMSFileObject(FileActivity):
    """Pydantic model for LMS file `object` property.

    Attributes:
        definition (dict): see LMSFileObjectDefinition.
    """

    definition: LMSFileObjectDefinition



"""LMS xAPI event definitions."""

from datetime import datetime

from ...selector import selector
from ..base.statements import BaseXapiStatement
from ..concepts.activity_types.audio import AudioActivity
from ..concepts.activity_types.scorm_profile import CourseActivity
from ..concepts.activity_types.tincan_vocabulary import DocumentActivity
from ..concepts.activity_types.video import VideoActivity
from ..concepts.verbs.adl_vocabulary import RegisteredVerb
from ..concepts.verbs.navy_common_reference_profile import AccessedVerb, UploadedVerb
from ..concepts.verbs.tincan_vocabulary import DownloadedVerb, UnregisteredVerb
from .contexts import (
    LMSCommonContext,
    LMSContext,
    LMSDownloadedAudioContext,
    LMSDownloadedVideoContext,
    LMSRegistrationContext,
)
from .objects import LMSFileObject, LMSPageObject


class BaseLMSStatement(BaseXapiStatement):
    """Pydantic model for LMS core statements.

    Attributes:
        timestamp (datetime): Timestamp of when the event occurred.
    """

    timestamp: datetime
    context: LMSContext


# Course registration
class LMSRegisteredCourse(BaseLMSStatement):
    """Pydantic model for LMS `registered to a course` statement.

    Example: John is registered to a course.

    Attributes:
        verb (dict): See RegisteredVerb.
        object (dict): See CourseActivity.
        context (dict): See LMSRegistrationContext.
    """

    __selector__ = selector(
        verb__id="http://adlnet.gov/expapi/verbs/registered",
        object__definition__type="http://adlnet.gov/expapi/activities/course",
    )

    verb: RegisteredVerb = RegisteredVerb()
    object: CourseActivity
    context: LMSRegistrationContext


class LMSUnregisteredCourse(BaseLMSStatement):
    """Pydantic model for LMS `unregistered to a course` statement.

    Example: John is unregistered to a course.

    Attributes:
        verb (dict): See UnregisteredVerb.
        object (dict): See CourseActivity.
        context (dict): See LMSRegistrationContext.
    """

    __selector__ = selector(
        verb__id="http://id.tincanapi.com/verb/unregistered",
        object__definition__type="http://adlnet.gov/expapi/activities/course",
    )

    verb: UnregisteredVerb = UnregisteredVerb()
    object: CourseActivity
    context: LMSRegistrationContext


# Page
class LMSAccessedPage(BaseLMSStatement):
    """Pydantic model for LMS `accessed a page` statement.

    Example: John has accessed a page of an LMS or a website.

    Attributes:
        verb (dict): See AccessedVerb.
        object (dict): See LMSPageObject.
    """

    __selector__ = selector(
        verb__id="https://w3id.org/xapi/netc/verbs/accessed",
        object__definition__type="https://w3id.org/xapi/acrossx/activities/webpage",
    )

    verb: AccessedVerb = AccessedVerb()
    object: LMSPageObject


# File
class LMSAccessedFile(BaseLMSStatement):
    """Pydantic model for LMS `accessed a file` statement.

    Example: John has accessed a file such as a pdf, doc, txt, ppt, xls, csv, etc...

    Attributes:
        verb (dict): See AccessedVerb.
        object (dict): See LMSFileObject.
    """

    __selector__ = selector(
        verb__id="https://w3id.org/xapi/netc/verbs/accessed",
        object__definition__type="http://activitystrea.ms/file",
    )

    verb: AccessedVerb = AccessedVerb()
    object: LMSFileObject


class LMSUploadedFile(BaseLMSStatement):
    """Pydantic model for LMS `uploaded a file` statement.

    Example: John has uploaded a file such as a pdf, doc, txt, ppt, xls, csv, etc...

    Attributes:
        verb (dict): See UploadedVerb.
        object (dict): See LMSFileObject.
        context (dict) See LMSCommonContext.
    """

    __selector__ = selector(
        verb__id="https://w3id.org/xapi/netc/verbs/uploaded",
        object__definition__type="http://activitystrea.ms/file",
    )

    verb: UploadedVerb = UploadedVerb()
    object: LMSFileObject
    context: LMSCommonContext


class LMSDownloadedFile(BaseLMSStatement):
    """Pydantic model for LMS `downloaded a file` statement.

    Example: John has uploaded a file such as a pdf, doc, txt, ppt, xls, csv, etc...

    Attributes:
        verb (dict): See DownloadedVerb.
        object (dict): See LMSFileObject.
        context (dict): see LMSCommonContext.
    """

    __selector__ = selector(
        verb__id="http://id.tincanapi.com/verb/downloaded",
        object__definition__type="http://activitystrea.ms/file",
    )

    verb: DownloadedVerb = DownloadedVerb()
    object: LMSFileObject
    context: LMSCommonContext


# Video
class LMSUploadedVideo(BaseLMSStatement):
    """Pydantic model for LMS `uploaded a video` statement.

    Example: John uploaded a video.

    Attributes:
        verb (dict): See UploadedVerb.
        object (dict): See VideoActivity.
        context (dict): see LMSCommonContext.
    """

    __selector__ = selector(
        object__definition__type="https://w3id.org/xapi/video/activity-type/video",
        verb__id="https://w3id.org/xapi/netc/verbs/uploaded",
    )

    object: VideoActivity
    verb: UploadedVerb = UploadedVerb()
    context: LMSCommonContext


class LMSDownloadedVideo(BaseLMSStatement):
    """Pydantic model for LMS `downloaded a video` statement.

    Example: John downloaded (rather than played) a video.

    Attributes:
        verb (dict): See DownloadedVerb.
        object (dict): See VideoActivity.
        context (dict): See LMSDownloadedVideoContext.
    """

    __selector__ = selector(
        object__definition__type="https://w3id.org/xapi/video/activity-type/video",
        verb__id="http://id.tincanapi.com/verb/downloaded",
    )

    object: VideoActivity
    verb: DownloadedVerb = DownloadedVerb()
    context: LMSDownloadedVideoContext


# Document
class LMSUploadedDocument(BaseLMSStatement):
    """Pydantic model for LMS `uploaded a document` statement.

    Example: John uploaded a document.

    Attributes:
        verb (dict): See UploadedVerb.
        object (dict): See DocumentActivity.
        context (dict): see LMSCommonContext.
    """

    __selector__ = selector(
        object__definition__type="http://id.tincanapi.com/activitytype/document",
        verb__id="https://w3id.org/xapi/netc/verbs/uploaded",
    )

    object: DocumentActivity
    verb: UploadedVerb = UploadedVerb()
    context: LMSCommonContext


class LMSDownloadedDocument(BaseLMSStatement):
    """Pydantic model for LMS `downloaded a document` statement.

    Example: John downloaded (rather than accessed or opened) a document.

    Attributes:
        verb (dict): See DownloadedVerb.
        object (dict): See DocumentActivity.
        context (dict): see LMSCommonContext.
    """

    __selector__ = selector(
        object__definition__type="http://id.tincanapi.com/activitytype/document",
        verb__id="http://id.tincanapi.com/verb/downloaded",
    )

    object: DocumentActivity
    verb: DownloadedVerb = DownloadedVerb()
    context: LMSCommonContext


# Audio
class LMSUploadedAudio(BaseLMSStatement):
    """Pydantic model for LMS `uploaded an audio` statement.

    Example: John uploaded an audio.

    Attributes:
        verb (dict): See UploadedVerb.
        object (dict): See AudioActivity.
        context (dict): see LMSCommonContext.
    """

    __selector__ = selector(
        object__definition__type="https://w3id.org/xapi/audio/activity-type/audio",
        verb__id="https://w3id.org/xapi/netc/verbs/uploaded",
    )

    object: AudioActivity
    verb: UploadedVerb = UploadedVerb()
    context: LMSCommonContext


class LMSDownloadedAudio(BaseLMSStatement):
    """Pydantic model for LMS `downloaded an audio` statement.

    Example: John downloaded (rather than played) an audio.

    Attributes:
        verb (dict): See DownloadedVerb.
        object (dict): See AudioActivity.
        context (dict): see LMSDownloadedAudioContext.
    """

    __selector__ = selector(
        object__definition__type="https://w3id.org/xapi/audio/activity-type/audio",
        verb__id="http://id.tincanapi.com/verb/downloaded",
    )

    object: AudioActivity
    verb: DownloadedVerb = DownloadedVerb()
    context: LMSDownloadedAudioContext
